import numpy as np
import matplotlib.pyplot as plt
from sklearn.linear_model import LinearRegression

def paretoplot(x, k):
    """
    Create a Pareto quantile plot for regularly varying variables.

    Parameters:
    -----------
    x : array-like
        The input data.

    k : int
        The number of largest values to consider for the Pareto plot.

    Returns:
    --------
    reg : LinearRegression
        The fitted linear regression model.
    """
    # Select the k largest values from the input data
    xEx = x[x >= np.sort(x)[::-1][k-1]]
    k = len(xEx)

    # Compute the logarithm of the sorted values
    Lxord = np.log(np.sort(xEx))

    # Compute the indices for the Pareto plot
    inds = 1 - (np.arange(1, k + 1)) / (k + 1)
    negLinds = -np.log(inds)

    # Create the Pareto plot
    plt.figure(figsize=(10, 6))
    plt.scatter(negLinds, Lxord, label='Data points')
    plt.title(f"Pareto Quantile Plot, {k} Largest Values")

    # Fit a linear regression model to the data
    reg = LinearRegression().fit(negLinds.reshape(-1, 1), Lxord)
    coef = [reg.intercept_, reg.coef_[0]]

    # Plot the fitted line
    plt.plot(negLinds, reg.predict(negLinds.reshape(-1, 1)), 'r-', linewidth=2, label='Fitted line')

    # Add labels and legend
    plt.xlabel("Negative Log Indices")
    plt.ylabel("Log of Sorted Values")
    plt.legend()
    plt.grid(True)
    plt.show()

    return reg

# Example usage:
# x = np.random.pareto(1, 1000) + 1  # Generate Pareto-distributed data
# paretoplot(x, 100)




from __future__ import division
# import random as rd
#import pdb
import numpy as np
import matplotlib.pyplot as plt
import warnings
from itertools import combinations
import networkx as nx
from sklearn.model_selection import KFold
from MLExtreme.utils.EVT_basics import rank_transform, round_signif
# from .ftclust_analysis import subfaces_list_to_matrix

def subfaces_list_to_matrix(subfaces, dimension):
    """
    Converts a list of subfaces into a binary matrix.

    Parameters:
    -----------
    - subfaces (list): List of subfaces.
    - dimension (int): Dimensionality of the ambient space.

    Returns:
    --------
    - np.ndarray: Binary matrix representation of subfaces.
    """
    num_subfaces = len(subfaces)

    if dimension is None:
        features = list({j for subface in subfaces for j in subface})
        dimension = int(max(features)) + 1

    matrix_subfaces = np.zeros((num_subfaces, dimension))

    for subface_index, subface in enumerate(subfaces):
        matrix_subfaces[subface_index, subface] = 1

    return matrix_subfaces  # [:, np.sum(matrix_subfaces, axis=0) > 0]


def remove_zero_rows(binary_array):
    """
    Removes rows that contain only zeros from a binary matrix.

    Parameters:
    -----------
    - binary_array (np.ndarray): A binary matrix from which to remove zero
        rows.

    Returns:
    --------
    - np.ndarray: The binary matrix with zero rows removed.
    """
    # Remove rows where the sum of elements is zero (i.e., all
    # elements are zero)
    return binary_array[np.sum(binary_array, axis=1) > 0]


def binary_large_features(std_data, radial_threshold, epsilon):
    """
    Filters extreme samples from the input `std_data`, returns a
    binary matrix of same number of rows as the number of extreme
    samples, where each row is a binary vector indicating which
    components of the extreme sample point are large. Here the radius
    of a sample point is its infinite norm and a sample point is
    extreme if its radius is greater than `radial_threshold'. Also a
    sample's component is 'large' means that it is greater than
    `epsilon * radial_threshold'.

    Parameters:
    -----------
    - std_data (np.ndarray): A data matrix. For meaningful usage,
      columns should be preliminarily standardized.
    - radial_threshold (float): The threshold value to compare against.
    - epsilon (float, optional): A tolerance parameter between 0 and 1
        for the radius threshold. If None, treated as 1. A value of
        zero will have the function declare as 'large' every
        components in extreme sample points. On the contrary, with
        `epsilon=1`, a component needs to exceed the radial threshold
        to be declared large.

    Returns:
    -------
    - np.ndarray: A binary matrix with values above the
    radius threshold, with zero rows removed.

    Note: For standardizing a raw data input, see mlx.rank_transform,
    mlx.rank_transform_test.
    """
    if epsilon is not None:
        # Create a binary matrix where values are above the radial
        # threshold rescaled by epsilon, after selecting samples which
        # radius exceeds the radial threshold.
        extreme_samples = std_data[np.max(std_data, axis=1) >=
                                   radial_threshold]
        binary_matrix = extreme_samples > radial_threshold * epsilon
    else:
        # Create a binary matrix which values are  above the
        # radius threshold
        binary_matrix = std_data >= radial_threshold

    # Convert the binary matrix to float type and remove zero rows
    return remove_zero_rows(binary_matrix.astype(int))


def estim_subfaces_mass(subfaces_list, X, threshold, epsilon,
                        standardize):
    """
    Estimates the mass of each subface in `subfaces_list` based on the
    data `X`.

    Parameters:
    - subfaces_list (list): List of subfaces.
    - X (np.ndarray): Input data.
    - threshold (float): Threshold for identifying extremes.
    - epsilon (float): Tolerance parameter for identifying large
        components.
    - standardize (bool): Whether to standardize the data.

    Returns:
    - list: Estimated masses for each subface.
    """
    if standardize:
        x_norm = rank_transform(X)
    else:
        x_norm = X
    x_bin = binary_large_features(x_norm, threshold, epsilon)
    dimension = np.shape(x_bin)[1]
    mass_list = []
    subfaces_matrix = subfaces_list_to_matrix(subfaces_list, dimension)

    for k in range(len(subfaces_list)):
        subface = subfaces_matrix[k]
        shared_features = np.dot(x_bin, subface.reshape(-1, 1))
        # (num_extremes, 1)
        num_features_samples = np.sum(x_bin, 1).reshape(-1, 1)
        # (num_extremes, 1)
        num_features_subface = np.sum(subface)  # int
        sample_superset_of_subface = num_features_subface == shared_features
        sample_subset_of_subface = num_features_samples == shared_features
        sample_equal_subface = sample_subset_of_subface * \
            sample_superset_of_subface
        counts_subface = np.sum(sample_equal_subface)
        mass = threshold * counts_subface / X.shape[0]
        mass_list.append(float(mass))

    return np.array(mass_list)


##################################################
# DAMEX functions
##################################################

def damex_0(binary_matrix, include_singletons):
    """
    Analyzes a binary matrix to determine the number of points (row) matching
    patterns in {0,1}^d where d is the number of columns in the matrix.

    Parameters:
    - binary_matrix (np.ndarray): A binary matrix where rows represent samples
        and columns represent features.
    - include_singletons (bool): Whether to include singletons in the analysis.

    Returns:
    - faces (list of lists): A list of subfaces of the infinite sphere,
        each represented as a list of feature indices.
    - mass (np.ndarray): An array indicating the number of rescaled
        samples observed within each subface.
    """
    # Set the minimum number of features per subface
    if include_singletons:
        min_features = 1
    else:
        min_features = 2

    # Sample size
    num_samples = binary_matrix.shape[0]
    # Number of features (columns) with value 1 for each
    # sample (row)
    num_features_per_sample = np.sum(binary_matrix, axis=1)

    # Dot product of the binary matrix with its
    # transpose: get shared features between samples.
    # Entry ij is the number of features shared by samples i and j.
    shared_features_matrix = np.dot(binary_matrix, binary_matrix.T)

    # Determine samples with exactly matching features
    # Entry ij is one if (binary) samples i and j are identical
    exact_match_matrix = (
        shared_features_matrix == num_features_per_sample) & (
            shared_features_matrix.T == num_features_per_sample).T

    # Set of samples not yet observed/assigned in any subface
    uncovered_samples = set(range(num_samples))

    # Dictionary to store the number of samples assigned to each subface
    subface_sample_count = {}

    # Iterate over each sample to identify subfaces
    for i in range(num_samples):
        # Find samples with exactly matching features
        matching_samples = list(np.nonzero(exact_match_matrix[i, :])[0])

        # If the current sample has not been assigned yet
        if i in uncovered_samples:
            # Mark these samples as assigned
            uncovered_samples -= set(matching_samples)

            # If the sample has more than one feature, record the
            # subface (add a key to the dictionary with value equal to
            # the number of samples in this subface. The key is the
            # index of the first record of the event that a sample
            # point belongs to the considered subface.
            if num_features_per_sample[i] >= min_features:
                subface_sample_count[i] = len(matching_samples)

    # Sort subfaces by the number of samples they cover, in descending order
    sorted_indices = np.argsort(list(subface_sample_count.values()))[::-1]

    # Create a (sorted) list of subfaces, each represented by the
    # indices of `1` features
    subfaces = [list(np.nonzero(
        binary_matrix[list(subface_sample_count)[i], :])[0])
                for i in sorted_indices]

    converted_subfaces = [[int(item) for item in sublist]
                          for sublist in subfaces]

    # Create an array of the number of samples covered by each subface
    counts = [list(subface_sample_count.values())[i] for i in sorted_indices]

    return converted_subfaces, np.array(counts)

def damex_fit(data, threshold, epsilon, min_counts, standardize,
              include_singletons):
    """
    Implements the methodology proposed in [1].

    Given a data matrix (n,d), damex provides the groups of features that are
    likely to be simultaneously large, while the other features are small. The
    function also returns an estimate of the limit measure associated with each
    subgroup.

    Parameters:
    - data (np.ndarray): A data matrix.
    - threshold (float): The radial threshold for selecting extreme
      samples.
    - epsilon (float): A scaling factor for the radial threshold.
    - min_counts (int): The minimum number of points required per face.
        Defaults to 0.
    - standardize (bool): Defaults to True. If True, the data matrix will be
      standardized using the rank transformation mlx.rank_transform. Meaningful
      usage requires that columns (features) follow a unit Pareto distribution
        or
      that rank-standardization has been applied previously with
      `standardize=False`, or that columns are possibly not standard and
      `standardize=True`.

    Returns:
    - list: A list of faces where each face has more than min_points points.

    References:
    [1] Goix, N., Sabourin, A., & Clémençon, S. (2017). Sparse representation
    of multivariate extremes with applications to anomaly detection.
    Journal of Multivariate Analysis, 161, 12-31.
    (The returned estimator is defined in
    Equation (3.3) from the reference)
    """
    # Standardize data if necessary
    if standardize:
        intern_data = rank_transform(data)
    else:
        intern_data = data

    # Generate binary matrix and determine faces and their masses
    binary_matrix = binary_large_features(intern_data, threshold,
                                          epsilon)
    faces, counts = damex_0(binary_matrix, include_singletons)
    n = intern_data.shape[0]

    limit_mass_estimator = threshold * counts / n
    id_large_mass = (counts >= min_counts)
    number_heavy_faces = np.sum(id_large_mass)
    truncated_faces = faces[:number_heavy_faces]
    truncated_mass_estimator = limit_mass_estimator[:number_heavy_faces]

    # Return faces with mass greater than or equal to min_points
    return truncated_faces, truncated_mass_estimator

# def damex_estim_subfaces_mass(subfaces_list, X, threshold, epsilon,
#                               standardize=True):
#     res = estim_subfaces_mass(subfaces_list, X, threshold, epsilon,
#                               standardize=standardize)
#     return res

##################################################
# # CLEF
##################################################

#############
# Clef algo #
#############

def clef_fit(X, radius, kappa_min, standardize, include_singletons):
    """
    Returns maximal faces such that kappa > kappa_min.

    Parameters:
    - X (np.ndarray): Input data.
    - radius (float): Radius threshold for selecting extreme samples.
    - kappa_min (float): Minimum kappa value.
    - standardize (bool): Whether to standardize the data.
    - include_singletons (bool): Whether to include singletons.

    Returns:
    - list: List of maximal faces.
    """
    if standardize:
        x_norm = rank_transform(X)
    else:
        x_norm = X
    x_bin = binary_large_features(x_norm, radius, epsilon=None)
    result = clef_fit0(x_bin, kappa_min, include_singletons)
    return result

def clef_fit0(x_bin, kappa_min, include_singletons):
    """
    Returns maximal faces such that kappa > kappa_min.

    Parameters:
    - x_bin (np.ndarray): Binary matrix of extreme samples.
    - kappa_min (float): Minimum kappa value.
    - include_singletons (bool): Whether to include singletons.

    Returns:
    - list: List of maximal faces.
    """
    faces_dict = find_faces(x_bin, kappa_min)
    faces = find_maximal_faces(faces_dict)

    if include_singletons:
        dim = np.shape(x_bin)[1]
        missed_singletons = cons_missing_singletons(faces, dim)
        for sing in missed_singletons:
            faces.append(sing)
    return faces

##################
# CLEF functions #
##################

def cons_missing_singletons(faces, d):
    """
    Constructs missing singletons from a list of faces.

    Parameters:
    - faces (list): List of faces.
    - d (int): Dimensionality of the space.

    Returns:
    - list: List of missing singletons.
    """
    # Create a set of all possible indices
    all_indices = set(range(d))
    # Iterate through each sublist in faces and remove its elements
    # from all_indices
    for sublist in faces:
        all_indices.difference_update(sublist)

    # Convert the remaining indices to a list and return
    list_indices = list(all_indices)
    return list([index] for index in list_indices)

def faces_init(x_bin, mu_0):
    """
    Returns faces of size 2 such that kappa > kappa_min.

    Parameters:
    - x_bin (np.ndarray): Binary matrix of extreme samples.
    - mu_0 (float): Minimum kappa value.

    Returns:
    - list: List of asymptotic pairs.
    """
    asymptotic_pair = []
    for (i, j) in combinations(range(x_bin.shape[1]), 2):
        pair_tmp = x_bin[:, [i, j]]
        one_out_of_two = np.sum(np.sum(pair_tmp, axis=1) > 0)
        two_on_two = np.sum(np.prod(pair_tmp, axis=1))
        if one_out_of_two > 0:
            proba = two_on_two / one_out_of_two
            if proba > mu_0:
                asymptotic_pair.append([i, j])

    return asymptotic_pair

def compute_beta(x_bin, face):
    """
    Computes the beta value for a given face.

    Parameters:
    - x_bin (np.ndarray): Binary matrix of extreme samples.
    - face (list): Face to evaluate.

    Returns:
    - int: Beta value.
    """
    return np.sum(np.sum(x_bin[:, face], axis=1) > len(face)-2)

def kappa(x_bin, face):
    """
    Returns the kappa value.

    kappa = #{i | for all j in face, X_ij=1} /
    #{i | at least |face|-1 j, X_ij=1}

    Parameters:
    - x_bin (np.ndarray): Binary matrix of extreme samples.
    - face (list): Face to evaluate.

    Returns:
    - float: Kappa value.
    """
    beta = compute_beta(x_bin, face)
    all_face = np.sum(np.prod(x_bin[:, face], axis=1))
    if beta == 0.:
        kap = 0.
    else:
        kap = all_face / float(beta)

    return kap

def find_faces(x_bin, kappa_min):
    """
    Returns all faces such that kappa > kappa_min.

    Parameters:
    - x_bin (np.ndarray): Binary matrix of extreme samples.
    - kappa_min (float): Minimum kappa value.

    Returns:
    - dict: Dictionary of faces.
    """
    dim = x_bin.shape[1]
    size = 2
    faces_dict = {}
    faces_dict[size] = faces_init(x_bin, kappa_min)
    # print('face size : nb faces')
    while len(faces_dict[size]) > size:
        # print(size, ':', len(faces_dict[size]))
        faces_dict[size + 1] = []
        faces_to_try = candidate_subfaces(faces_dict[size], size, dim)
        if faces_to_try:
            for face in faces_to_try:
                if kappa(x_bin, face) > kappa_min:
                    faces_dict[size + 1].append(face)
        size += 1

    return faces_dict

def find_maximal_faces(faces_dict, lst=True):
    """
    Returns inclusion-wise maximal faces.

    Parameters:
    - faces_dict (dict): Dictionary of faces.
    - lst (bool): Whether to return a list of faces.

    Returns:
    - list: List of maximal faces.
    """
    # k = len(faces_dict.keys()) + 1
    if len(faces_dict) == 0:
        return []
    list_keys = []
    for k in faces_dict.keys():
        list_keys.append(k)
    list_keys.sort(reverse=True)
    k = list_keys[0]
    maximal_faces = [faces_dict[k]]
    faces_used = list(map(set, faces_dict[k]))
    # for i in list_keys[:-1]:  ##range(1, k - 1):
    for j in list_keys[1:]:
        face_tmp = list(map(set, faces_dict[j]))
        for face in faces_dict[j]:
            for face_test in faces_used:
                if len(set(face) & face_test) == j:
                    face_tmp.remove(set(face))
                    break
        maximal_faces.append(list(map(list, face_tmp)))
        faces_used = faces_used + face_tmp
    maximal_faces = maximal_faces[::-1]
    if lst:
        maximal_faces = [face for faces_ in maximal_faces
                         for face in faces_]

    return maximal_faces

##################
# Generate Candidate Subfaces of Increased Size at Each CLEF Iteration
##################

def make_graph(subfaces, size, dimension):
    """
    Creates a graph where nodes represent subfaces and edges exist if subfaces
    differ by at most one feature.

    It is the main building block of the `candidate_sufaces' function
    used in CLEF and variants [1,2].

    Parameters:
    - subfaces (list): List of subfaces.
    - size (int): Size of the subfaces.
    - dimension (int): Dimensionality of the subfaces.

    Returns:
    - nx.Graph: Graph representation of subfaces.

    References
    -----------

    [1] Chiapino, M., & Sabourin, A. (2016,September). Feature clustering
    for extreme events analysis, with application to extreme stream-flow data.
    In International workshop on new frontiers in mining complex patterns
    (pp. 132-147). Cham: Springer International Publishing.

    [2] Chiapino, M., Sabourin, A., & Segers, J. (2019). Identifying groups of
    variables with the potential of being large simultaneously.
    Extremes, 22, 193-222.
    """
    vector_subfaces = subfaces_list_to_matrix(subfaces, dimension)
    num_subfaces = len(vector_subfaces)
    graph = nx.Graph()
    nodes = range(num_subfaces)
    graph.add_nodes_from(nodes)
    edges = np.nonzero(np.triu(
        np.dot(vector_subfaces, vector_subfaces.T) == size - 1))
    graph.add_edges_from([(edges[0][i], edges[1][i])
                          for i in range(len(edges[0]))])

    return graph

def candidate_subfaces(subfaces, size, dimension):
    """
    Generates a list A_{s+1} of candidate subfaces of size s+1
    from a list A_s = `subfaces' of subfaces of size s, with `s=size`.
    Candidate subfaces are all subfaces of size s+1 which are
    supersets of each of subface of size s in the current list
    `subfaces'.

    This is a key step in CLEF algorithm [1] and variants [2]

    Parameters:
    - subfaces (list): List of subfaces.
    - size (int): Size of the subfaces.
    - dimension (int): Dimensionality of the subfaces.

    Returns:
    - list: List of candidate subfaces.

    References
    -----------

    [1] Chiapino, M., & Sabourin, A. (2016,September). Feature clustering
    for extreme events analysis, with application to extreme stream-flow data.
    In International workshop on new frontiers in mining complex patterns
    (pp. 132-147). Cham: Springer International Publishing.

    [2] Chiapino, M., Sabourin, A., & Segers, J. (2019). Identifying groups of
    variables with the potential of being large simultaneously.
    Extremes, 22, 193-222.
    """
    graph = make_graph(subfaces, size, dimension)
    candidate_subfaces_list = []
    cliques = list(nx.find_cliques(graph))
    indices_to_try = np.nonzero(np.array(
        list(map(len, cliques))) == size + 1)[0]

    for index in indices_to_try:
        clique_features = set([])
        for subface_index in range(len(cliques[index])):
            clique_features = clique_features | set(
                subfaces[cliques[index][subface_index]])
        clique_features = list(clique_features)
        if len(clique_features) == size + 1:
            candidate_subfaces_list.append(clique_features)

    return candidate_subfaces_list

############################
# Subfaces frequency analysis #
############################

# def rho_value(binary_matrix, subface, k):
#     """
#     Calculates the rho value of a subface.
#     (notation r_a(1) in [1], where a is the subface)

#     Args:
#     - binary_matrix (np.ndarray): Binary matrix.
#     - subface (list): Subface to evaluate.
#     - k (int): Number of samples.

#     Returns:
#     - float: Rho value.

#     References
#     -----------

#     [1] Chiapino, M., Sabourin, A., & Segers, J. (2019).
#       Identifying groups of variables with the potential of being
#        large simultaneously. Extremes, 22, 193-222.

#     """
#     return np.sum(np.sum(
#         binary_matrix[:, subface], axis=1) == len(subface)) / float(k)

def partial_matrix(base_matrix, partial_matrix, column_index):
    """
    Replaces the column_index column of base_matrix with the corresponding
    column from partial_matrix.

    Used in asymptotic variants of CLEF [1] to compute partial derivatives
    of the 'kappa' and the 'r' functions

    Parameters:
    - base_matrix (np.ndarray): Base matrix.
    - partial_matrix (np.ndarray): Partial matrix.
    - column_index (int): Index of the column to replace.

    Returns:
    - np.ndarray: Modified matrix.

    References
    -----------

    [1] Chiapino, M., Sabourin, A., & Segers, J. (2019). Identifying groups of
    variables with the potential of being large simultaneously.
    Extremes, 22, 193-222.
    """
    matrix_copy = np.copy(base_matrix)
    matrix_copy[:, column_index] = partial_matrix[:, column_index]

    return matrix_copy

def clef_estim_subfaces_mass(subfaces_list, X, threshold,
                             standardize):
    """
    Estimates the mass of each subface in `subfaces_list` based on the
    data `X`.

    Parameters:
    - subfaces_list (list): List of subfaces.
    - X (np.ndarray): Input data.
    - threshold (float): Threshold for identifying extremes.
    - standardize (bool): Whether to standardize the data.

    Returns:
    - list: Estimated masses for each subface.
    """
    res = estim_subfaces_mass(subfaces_list, X, threshold, epsilon=None,
                              standardize=standardize)
    return res


##################################################
# Feature clusters analysis: unit deviance, total deviance,
# dispersion model ...
# previously in ftclust_analysis
##################################################

def unit_deviance(subface, subfaces_matrix):
    """Calculates unit deviances `d` between one subface and a matrix of
    (binary encoded) subfaces.

    The unit deviance `d` between any two subfaces is defined as the
    ratio between the cardinalities of the symmetric difference and
    the union of the two sets.

    This is the key ingredient of the unsupervised performance
    criterion s(list_of_subsets, binary_encoded_new_point) proposed in [1]

    In the present implementation it is the basic building block of
    the total deviance, which serves as convenient unsupervised
    criterion for parameter tuning.

    Parameters:
    -----------
    subface : np.ndarray)
        Single subface vector (binary vector), typically
        a test vector.
    - subfaces_matrix : np.ndarray)
        Matrix of reference subface vectors
      (binary entries) (typically, the ones estimated by damex or clef).

    Returns:
    --------
    - np.ndarray: Array of distances.

    References:
    -----------

    [1] Chiapino, M., Sabourin, A., & Segers, J. (2019). Identifying groups of
    variables with the potential of being large simultaneously.
    Extremes, 22, 193-222.

    """
    result = (np.sum(abs(subface - subfaces_matrix), axis=1) /
              np.sum(subface + subfaces_matrix > 0, axis=1))

    return result

def total_deviance_binary_matrices(subfaces_matrix, masses,
                                   subfaces_reference_matrix,
                                   reference_masses, 
                                   rate):
    """Computes the total deviance between all rows of subfaces_matrix
    and the matrix subfaces_reference_matrix, and returns a normalized
    version of it. If `mass` is not provided, the normalization is
    done by dividing the total deviance by the sample size (i.e. the
    number of rows in subfaces_matrix).  Alternatively, if `mass` is
    provided, the contribution of each row is weighted by the relative
    weight of the corresponding entry in `mass`.

    The total deviance is relative to a dispersion_model. It is the
    (normalized) negative log-likelihood of the rows in
    subfaces_matrix, within a mixture model defined in the set of
    subfaces, as follows:

        - each row in subfaces_reference_matrix is a 'center of mass'
    
        - the references_masses are the (unnormalized) weights of the mixture

    Given that row_i in subfaces_matrix was generated by the mixture
            component reference_row_j in subfaces_reference_matrix,
            the unit deviance (see `unit_deviance`) between row_i and
            reference_row_j follows an exponential distribution with
            rate `rate`. The likelihood term disregards any necessary
            normalizing constants, such as those accounting for the
            truncation of the exponential model or the combinatorial
            number of subfaces at a given set distance from
            reference_row_j, which would be necessary in principle to
            define a proper dispersion model `à la Jorgensen'.

    In unsupervised usages:
        - subfaces_reference_matrix is typically issued from prior
    estimation from DAMEX or CLEF
        - subfaces_matrix is typically a dataset

    In supervised usage:
        Both subfaces_matrix and subfaces_reference matrix are issued from a
        list of true or estimated subfaces and the goal is to assess the
        'distance' between the two.

    The lower, the better fit.

    Parameters:
    - subfaces_matrix (np.ndarray): Matrix of subfaces.
    - masses (list or np.ndarray): Masses associated with subfaces.
    - subfaces_reference_matrix (np.ndarray): Reference matrix of subfaces.
    - reference_masses (list or np.ndarray): Masses associated with reference
      subfaces.
    - rate (float, >0): Rate parameter for deviance calculation.

    Returns:
    - float: Average aggregated set distance.

    """
    n_subfaces = np.shape(subfaces_matrix)[0]
    if n_subfaces == 0:
        return 0
    neg_log_likelihood = 10 * np.ones(n_subfaces)
    if masses is None:
        weights = np.ones(n_subfaces)/n_subfaces
    else:
        if isinstance(masses, list):
            masses = np.array(masses)
        weights = masses / np.sum(masses)
    if reference_masses is None:
        warnings.warn('No reference masses provided. \
        Using uniform weights by default.')
        nsub_ref = subfaces_reference_matrix.shape[0]
        reference_masses = np.ones(nsub_ref)/nsub_ref

    if isinstance(reference_masses, list):
        reference_masses = np.array(reference_masses)

    total_reference_mass = np.sum(reference_masses)
    ref_weights = reference_masses/total_reference_mass \
        if total_reference_mass > 0 else reference_masses

    for i in range(n_subfaces):
        unit_deviance_vect = unit_deviance(subfaces_matrix[i],
                                           subfaces_reference_matrix)
        mixture_likelihood = np.sum(ref_weights *
                                    np.exp(- rate * unit_deviance_vect))
        
        if mixture_likelihood > 0:
            neg_log_likelihood[i] = - np.log(mixture_likelihood)
        else:
            # Handle the case where mixture_likelihood is zero
            neg_log_likelihood[i] = np.inf
            
    # return negative log-likelihood, normalized by sample size or
    # sample's weights
            
    return np.sum(neg_log_likelihood * weights)


##################################
# For unsupervised evaluation: unit deviance, total deviance in a
# dispersion model###
#################################

def total_deviance(subfaces_list, masses, std_data, threshold,
                   include_singletons_test, epsilon, rate):
    """Calculates the total deviance for the estimated parameter
    (list of subfaces, list of masses), evaluated on  the multivariate
    peaks-over-threshold of a preliminary standardized dataset
    std_data.  The total deviance is normalized by the number of
    extreme samples in std_data

    Parameters:
    - subfaces (list): List of subfaces.

    - masses (array): associated masses

    - std_data (np.ndarray): Standardized data.

    - threshold (float): Radius for selection of extremes in std_data

    - include_singletons_test: if False, disregards events where a
      single feature is large (suitable for analysising concomittant
      events)

    Returns:
    - float: Average distance.

    """
    if len(subfaces_list) == 0:
        return float('inf')

    subfaces_matrix = subfaces_list_to_matrix(subfaces_list,
                                              dimension=std_data.shape[1])
    binary_data = binary_large_features(std_data, threshold,
                                        epsilon=epsilon)

    if isinstance(masses, list):
        masses = np.array(masses)

    if not include_singletons_test:
        id_keep_data = np.where(np.sum(binary_data, axis=1) >= 2)[0]
        binary_data = binary_data[id_keep_data]

    if masses is None:
        warnings.warn('No masses given for subfaces_list.\
        Estimating on test dataset and proceeding ...')
        masses = estim_subfaces_mass(subfaces_list, std_data, threshold,
                                     epsilon, standardize=False)

    Deviance = total_deviance_binary_matrices(
        subfaces_matrix=binary_data, masses=None,
        subfaces_reference_matrix=subfaces_matrix,
        reference_masses=masses, rate=rate)

    return Deviance


#################################
# ##### previously in damex
#################################

def list_to_dict(faces_list, mass_list):
    """
    Converts a list of faces into a dictionary where keys are face
    sizes and values are lists of faces, orderted by sizes.

    Useful for inspection of large lists of subfaces.

    Parameters:
    - faces_list (list of lists): A list where each element is a list
      of points representing a face.
    - mass_list (list): Optional. If provided, the function
      returns a dictionary with the same keys as faces_dict and the values are
      the masses associated with the corresponding subfaces in faces_dict.

    Returns:
    - faces_dict: A dictionary where the key is the size of the face and the
      value is a list of faces of that size.
    - mass_dict: A dictionary with the same keys as faces_dict and the values
      are the masses associated with the corresponding subfaces in faces_dict.
    """
    if len(faces_list) == 0:
        return {}, {}
    # Initialize dictionary with sizes ranging from 1 to the maximum face size
    faces_dict = {size: [] for size in range(1, max(map(len, faces_list)) + 1)}
    mass_dict = {size: [] for size in range(1, max(map(len, faces_list)) + 1)}
    # Populate the dictionary with faces based on their sizes
    for index, face in enumerate(faces_list):
        faces_dict[len(face)].append(face)
        if mass_list is not None:
            mass_dict[len(face)].append(round_signif(mass_list[index], 2))

    return faces_dict, mass_dict

def ftclust_cross_validate(X, standardize, algo, tolerance,
                           min_counts, use_max_subfaces,
                           thresh_train,
                           thresh_test, include_singletons_train,
                           include_singletons_test, rate, cv,
                           random_state):
    """
    Performs cross-validation for feature clustering.

    Parameters:
    - X (np.ndarray): Input data.
    - standardize (bool): Whether to standardize the data.
    - algo (str): Algorithm to use ('damex' or 'clef').
    - tolerance (float): Tolerance level for clustering.
    - min_counts (int): Minimum number of points required to form a cluster.
    - use_max_subfaces (bool): Whether to use maximal subfaces.
    - thresh_train (float): Threshold for training data.
    - thresh_test (float): Threshold for test data.
    - include_singletons_train (bool): Whether to include singletons in
      training.
    - include_singletons_test (bool): Whether to include singletons in
      testing.
    - rate (float, >0): Rate parameter for deviance calculation.
    - cv (int): Number of cross-validation folds.
    - random_state (int): Random seed for reproducibility.

    Returns:
    - np.ndarray: Cross-validated deviance scores.
    """
    if standardize:
        Xt = rank_transform(X)
    else:
        Xt = X

    scores = []
    Norm_Xt = np.max(Xt, axis=1)

    if thresh_train is None:
        thresh_train = np.quantile(Norm_Xt, (1 - 1 / np.sqrt(len(Norm_Xt))))

    if thresh_test is None:
        thresh_test = thresh_train

    # which data are considered extreme for training and testing:
    # logical mask vectors of size len(y)
    id_extreme_train = (Norm_Xt >= thresh_train)
    id_extreme_predict = (Norm_Xt >= thresh_test)

    # K-fold train/test indices
    kf = KFold(n_splits=cv, shuffle=True, random_state=random_state)
    # CV loop
    for train_index, test_index in kf.split(X):
        size_train_ex = np.sum(id_extreme_train[train_index])
        size_predict_ex = np.sum(id_extreme_predict[test_index])
        if size_train_ex <= 2 or size_predict_ex <= 0:
            continue
        # Split the data into training and testing sets
        X_train, X_test = Xt[train_index], Xt[test_index]
        # total_mass_train = thresh_train * size_train_ex/len(train_index)

        #  fit the model
        if algo == 'clef':
            Subfaces = clef_fit(X_train, thresh_train, kappa_min=tolerance,
                                standardize=False,
                                include_singletons=include_singletons_train)
            Masses = clef_estim_subfaces_mass(Subfaces, X_train,
                                              thresh_train,
                                              standardize=False)
            epsilon_val = None
            # include_singletons_test = False
            # ## include_singletons_test is hard set to False
            # # here because  setting it to True has undesirable behaviour,
            # # namely it always selects a large kappa.

        elif algo == 'damex':
            # min_counts = kwargs.pop('min_counts', 0)
            faces_i, mass_i = damex_fit(
                X_train, thresh_train, epsilon=tolerance,
                min_counts=min_counts, standardize=False,
                include_singletons=include_singletons_train
            )
            Subfaces = faces_i
            Masses = mass_i

            if use_max_subfaces:
                faces_dict, _ = list_to_dict(faces_i, mass_i)
                Subfaces = find_maximal_faces(faces_dict, lst=True)
                Masses = estim_subfaces_mass(Subfaces, X_train,
                                             thresh_train,
                                             epsilon=tolerance,
                                             standardize=False)

            epsilon_val = tolerance
        else:
            raise ValueError("algo must be either`damex` or `clef`")

        clust_error = total_deviance(
            Subfaces, Masses,  X_test, thresh_test,
            include_singletons_test=include_singletons_test,
            epsilon=epsilon_val,
            rate=rate)

        scores.append(clust_error)

    return np.array(scores)

# thresh_train=None,
# thresh_test=None, include_singletons_train=False,
# include_singletons_test=False,  **kwargs):


def ftclust_choose_tolerance_cv(tolerance_grid, X, standardize, algo,
                                unstable_tol_max, min_counts,
                                use_max_subfaces,
                                thresh_train, thresh_test,
                                include_singletons_train,
                                include_singletons_test, rate, cv,
                                random_state, plot=False):
    """
    Chooses the optimal tolerance value based on cross-validation.

    Parameters:
    - tolerance_grid (list): List of tolerance values to test.
    - X (np.ndarray): Input data.
    - standardize (bool): Whether to standardize the data.
    - algo (str): Algorithm to use ('damex' or 'clef').
    - unstable_tol_max (float): Maximum tolerance value for unstable solutions.
    - min_counts (int): Minimum number of points required to form a cluster.
    - use_max_subfaces (bool): Whether to use maximal subfaces.
    - thresh_train (float): Threshold for training data.
    - thresh_test (float): Threshold for test data.
    - include_singletons_train (bool): Whether to include singletons in
      training.
    - include_singletons_test (bool): Whether to include singletons in
      testing.
    - rate (float, >0): Rate parameter for deviance calculation.
    - cv (int): Number of cross-validation folds.
    - random_state (int): Random seed for reproducibility.
    - plot (bool): Whether to plot the CV deviance values.

    Returns:
    - float: Selected tolerance value.
    - float: Deviance value for the selected tolerance.
    - np.ndarray: Vector of CV deviance values.
    """
    ntol = len(tolerance_grid)
    cv_error_vect = np.zeros(ntol)
    for i in range(ntol):
        cv_scores = ftclust_cross_validate(
            X, standardize, algo, tolerance=tolerance_grid[i],
            min_counts=min_counts,
            use_max_subfaces=use_max_subfaces,
            thresh_train=thresh_train,
            thresh_test=thresh_test,
            include_singletons_train=include_singletons_train,
            include_singletons_test=include_singletons_test,
            rate=rate,
            cv=cv,
            random_state=random_state
            )

        cv_error_vect[i] = np.mean(cv_scores)

    i_maxerr = np.argmax(cv_error_vect[tolerance_grid < unstable_tol_max])
    tol_maxerr = tolerance_grid[i_maxerr]
    i_mask = tolerance_grid <= tol_maxerr
    i_cv = np.argmin(cv_error_vect +
                     (1e+23) * i_mask)
    tol_cv = tolerance_grid[i_cv]
    error_tol_cv = cv_error_vect[i_cv]
    if plot:
        plt.scatter(tolerance_grid, cv_error_vect, c='gray', label='CV error')
        plt.plot([tol_cv, tol_cv], [0, error_tol_cv], c='red',
                 label='selected value')
        plt.grid()
        plt.legend()
        if algo == 'damex':
            plt.title("DAMEX: clustering pseudo-deviance, K-fold CV scores")
        else:
            plt.title("CLEF: clustering pseudo-deviance, K-fold CV scores")
        plt.show()
    return tol_cv, error_tol_cv, cv_error_vect




import numpy as np
import matplotlib.pyplot as plt
from copy import deepcopy
from . import utilities as ut
from MLExtreme.utils.EVT_basics import rank_transform, round_signif
# import pdb

# ################
# ## DAMEX CLASS
# ################


class damex:
    """
    An unsupervised feature clustering model to learn "which subsets of features of a random vector
    are likely to be simultaneously large, while the others are small".

    This class provides methods to fit the model to data, calculate deviance, compute the Akaike
    Information Criterion (AIC), select optimal epsilon values based on AIC or cross-validation,
    and calculate the deviance between estimated and true subfaces.

    Parameters:
    -----------
    epsilon : float, default=0.1
        Tolerance level for clustering.
    min_counts : int, default=0
        Minimum number of points required to form a cluster.
    thresh_train : float, optional
        Threshold for training data.
    thresh_test : float, optional
        Threshold for test data.
    include_singletons : bool, default=False
        Whether to include singletons in training and testing.
    rate : float, default=10
        Rate parameter for the dispersion model on subfaces.

    Attributes:
    -----------
    subfaces : list
        Identified subfaces.
    masses : array-like
        Masses associated with subfaces.
    total_mass : float
        Total mass of extremes.
    number_extremes : int
        Number of extreme points.
    dimension : int
        Dimensionality of the data.
    """

    def __init__(
        self,
        epsilon=0.1,
        min_counts=0,
        thresh_train=None,
        thresh_test=None,
        include_singletons=False,
        rate=10,
    ):
        """
        Initialize the DAMEX model with specified parameters.

        Parameters:
        -----------
        epsilon : float, default=0.1
            Tolerance level for clustering.
        min_counts : int, default=0
            Minimum number of points required to form a cluster.
        thresh_train : float, optional
            Threshold for training data.
        thresh_test : float, optional
            Threshold for test data.
        include_singletons : bool, default=False
            Whether to include singletons in training and testing.
        rate : float, default=10
            Rate parameter for the dispersion model on subfaces.
        """
        self.epsilon = epsilon
        self.min_counts = min_counts
        self.thresh_train = thresh_train
        self.thresh_test = thresh_test
        self.include_singletons = include_singletons
        self.rate = rate
        self.subfaces = None  # Identified subfaces
        self.masses = None  # Masses associated with subfaces
        self.total_mass = None  # Total mass of extremes
        self.number_extremes = None  # Number of extreme points
        self.dimension = None  # Dimensionality of the data

    def fit(self, X, epsilon=None, standardize=True):
        """
        Fit the DAMEX model to the data.

        Parameters:
        -----------
        X : np.ndarray
            Input data.
        epsilon : float, optional
            Tolerance level for clustering. If None, the instance's attribute will be used.
        standardize : bool, default=True
            Whether to standardize the data.

        Returns:
        --------
        Subfaces : list
            Identified subfaces.
        Masses : np.ndarray
            Masses associated with subfaces.
        """
        # Update instance attributes with optional arguments passed
        if epsilon is not None:
            self.epsilon = epsilon

        # Record dimension
        self.dimension = np.shape(X)[1]

        # Standardize X if needed
        Xt = rank_transform(X) if standardize else X
        norm_Xt = np.max(Xt, axis=1)

        # Set the training threshold if not provided
        if self.thresh_train is None:
            self.thresh_train = np.quantile(norm_Xt, 1 - 1 / np.sqrt(len(norm_Xt)))

        # Update total mass accordingly
        self.number_extremes = np.sum(norm_Xt > self.thresh_train)
        self.total_mass = self.thresh_train * self.number_extremes / len(norm_Xt)

        # Fit the model
        Subfaces, Masses = ut.damex_fit(
            Xt,
            self.thresh_train,
            self.epsilon,
            self.min_counts,
            standardize=False,
            include_singletons=self.include_singletons,
        )

        # Update instance's attributes
        self.subfaces = Subfaces
        self.masses = Masses
        return Subfaces, Masses

    def deviance(self, Xtest, thresh_test=None, standardize=False):
        """
        Calculate the deviance of the model on test data.

        For more details see :func:`clef.deviance`

        Parameters:
        -----------
        Xtest : np.ndarray
            Test data.
        thresh_test : float, optional
            Threshold for identifying extremes in the test set. If None, the instance's attribute will be used.
        standardize : bool, default=False
            Whether to standardize the data.

        Returns:
        --------
        float
            Deviance value.
        """
        if self.subfaces is None:
            raise RuntimeError(
                "The model has not been fitted yet. Call 'fit' "
                "with appropriate arguments before using this "
                "method."
            )

        Xt = rank_transform(Xtest) if standardize else Xtest

        # update test threshold if passed as argument
        if thresh_test is not None:
            self.thresh_test = thresh_test
        # set thresh_test to thresh_train if not previously specified
        if self.thresh_test is None:
            self.thresh_test = self.thresh_train

        Subfaces = self.subfaces
        Masses = self.masses

        negative_pseudo_lkl = ut.total_deviance(
            Subfaces,
            Masses,
            Xt,
            self.thresh_test,
            self.include_singletons,
            self.epsilon,
            rate=self.rate,
        )

        return 2 * negative_pseudo_lkl

    def get_AIC(self, Xtrain, standardize=True):
        """
        Calculate the Akaike Information Criterion (AIC) for the model.

        Parameters:
        -----------
        Xtrain : np.ndarray
            Training data.
        standardize : bool, default=True
            Whether to standardize the data.

        Returns:
        --------
        float
            AIC value.
        """
        if self.masses is None:
            raise RuntimeError("Fit the model before computing the AIC")

        Masses = self.masses
        intern_deviance = self.deviance(Xtrain, self.thresh_train, standardize)
        return intern_deviance + 2 * len(Masses) / self.number_extremes

    def select_epsilon_AIC(
        self,
        grid,
        X,
        standardize=True,
        unstable_eps_max=0.05,
        plot=False,
        update_epsilon=False,
    ):
        """
        Select the optimal epsilon value based on AIC.

        Parameters:
        -----------
        grid : list
            Grid of epsilon values to test.
        X : np.ndarray
            Input data.
        standardize : bool, default=True
            Whether to standardize the data.
        unstable_eps_max : float, default=0.05
            Maximum epsilon value for unstable solutions.
        plot : bool, default=False
            Whether to plot the AIC values.
        update_epsilon : bool, default=False
            Whether to update the model's epsilon value.

        Returns:
        --------
        eps_select_aic : float
            Selected epsilon value.
        aic_opt : float
            Optimal AIC value.
        vect_aic : np.ndarray
            Vector of AIC values.
        """
        old_epsilon = deepcopy(self.epsilon)
        Xt = rank_transform(X) if standardize else X
        ntests = len(grid)
        vect_aic = np.zeros(ntests)

        for counter, eps in enumerate(grid):
            _, _ = self.fit(Xt, eps, standardize=False)
            vect_aic[counter] = self.get_AIC(Xt, standardize=False)

        i_maxerr = np.argmax(vect_aic[grid < unstable_eps_max])
        eps_maxerr = grid[i_maxerr]
        i_mask = grid <= eps_maxerr
        i_minAIC = np.argmin(vect_aic + (1e23) * i_mask)
        eps_select_aic = grid[i_minAIC]
        aic_opt = vect_aic[i_minAIC]

        if plot:
            print(
                f"DAMEX: selected epsilon with AIC criterion "
                f"{round_signif(eps_select_aic, 2)}"
            )
            plt.figure(figsize=(10, 5))
            plt.xlabel("epsilon")
            plt.ylabel("AIC")
            plt.title("DAMEX- AIC versus epsilon")
            plt.scatter(grid, vect_aic, c="gray", label="AIC")
            plt.plot([eps_select_aic, eps_select_aic], [0, max(vect_aic)], c="red")
            plt.grid(True)
            plt.show()

        # Reset instance's epsilon value with initial or selected value
        if update_epsilon:
            self.epsilon = eps_select_aic
        else:
            self.epsilon = old_epsilon

        # Re-fit the model with initial or final epsilon
        _, _ = self.fit(Xt, self.epsilon, False)

        return eps_select_aic, aic_opt, vect_aic

    def deviance_CV(
        self,
        X,
        standardize=True,
        epsilon=None,
        thresh_test=None,
        cv=5,
        random_state=None,
    ):
        """
        Estimate the expected deviance using cross-validation.

        Parameters:
        -----------
        X : np.ndarray
            Input data.
        standardize : bool, default=True
            Whether to standardize the data.
        epsilon : float, optional
            Tolerance level for clustering. If None, the instance's attribute will be used.
        thresh_test : float, optional
            Radial threshold for test sets. If None, the instance's attribute will be used.
        cv : int, default=5
            Number of cross-validation folds.
        random_state : int, optional
            Random seed for reproducibility.

        Returns:
        --------
        np.ndarray
            Cross-validated deviance scores.
        """
        Xt = rank_transform(X) if standardize else X

        # Update instance's attributes if passed as arguments
        if epsilon is not None:
            self.epsilon = epsilon
        if thresh_test is not None:
            self.thresh_test = thresh_test
        if self.thresh_test is None:
            self.thresh_test = self.thresh_train

        cv_neglkl_scores = ut.ftclust_cross_validate(
            Xt,
            standardize=False,
            algo="damex",
            tolerance=self.epsilon,
            min_counts=self.min_counts,
            use_max_subfaces=False,
            thresh_train=self.thresh_train,
            thresh_test=self.thresh_test,
            include_singletons_train=self.include_singletons,
            include_singletons_test=self.include_singletons,
            rate=self.rate,
            cv=cv,
            random_state=random_state,
        )

        return 2 * cv_neglkl_scores

    def select_epsilon_CV(
        self,
        grid,
        X,
        standardize=True,
        update_epsilon=False,
        unstable_tol_max=0.05,
        thresh_test=None,
        cv=5,
        random_state=None,
        plot=False,
    ):
        """
        Select the optimal epsilon value based on cross-validation.

        Parameters:
        -----------
        grid : list
            Grid of epsilon values to test.
        X : np.ndarray
            Input data.
        standardize : bool, default=True
            Whether to standardize the data.
        update_epsilon : bool, default=False
            Whether to update the model's epsilon value.
        unstable_tol_max : float, default=0.05
            Maximum tolerance value for unstable solutions.
        cv : int, default=5
            Number of cross-validation folds.
        random_state : int, optional
            Random seed for reproducibility.
        plot : bool, default=False
            Whether to plot the CV deviance values.

        Returns:
        --------
        tol_cv : float
            Selected epsilon value.
        deviance_tol_cv : float
            Deviance value for the selected epsilon.
        cv_deviance_vect : np.ndarray
            Vector of CV deviance values.
        """
        Xt = rank_transform(X) if standardize else X
        epsilon_old = deepcopy(self.epsilon)
        ntol = len(grid)
        cv_deviance_vect = np.zeros(ntol)

        for i, eps in enumerate(grid):
            cv_scores = self.deviance_CV(Xt, False, eps, thresh_test, cv, random_state)
            cv_deviance_vect[i] = np.mean(cv_scores)

        i_maxerr = np.argmax(cv_deviance_vect[grid < unstable_tol_max])
        tol_maxerr = grid[i_maxerr]
        i_mask = grid <= tol_maxerr
        i_cv = np.argmin(cv_deviance_vect + (1e23) * i_mask)
        tol_cv = grid[i_cv]
        deviance_tol_cv = cv_deviance_vect[i_cv]

        if plot:
            print(
                f"DAMEX: selected epsilon with CV criterion {round_signif(tol_cv, 2)}"
            )
            plt.scatter(grid, cv_deviance_vect, c="gray", label="CV deviance")
            plt.plot(
                [tol_cv, tol_cv], [0, deviance_tol_cv], c="red", label="selected value"
            )
            plt.grid()
            plt.legend()
            plt.title("DAMEX: clustering pseudo-deviance, K-fold CV scores")
            plt.show()

        if update_epsilon:
            self.epsilon = tol_cv
        else:
            self.epsilon = epsilon_old

        # Re-fit the model with final epsilon
        _, _ = self.fit(Xt, False)

        return tol_cv, deviance_tol_cv, cv_deviance_vect

    def deviance_to_true(self, subfaces_true, weights_true):
        """
        Calculate the deviance between the estimated subfaces and the true subfaces.

        Parameters:
        -----------
        subfaces_true : list
            True subfaces.
        weights_true : list
            Weights of the true subfaces.

        Returns:
        --------
        est_to_truth : float
            Deviance from estimated to true subfaces.
        truth_to_est : float
            Deviance from true to estimated subfaces.
        """
        if self.subfaces is None:
            raise RuntimeError("DAMEX has not been fitted yet")
        Subfaces = self.subfaces
        Masses = self.masses

        if isinstance(Masses, list):
            Masses = np.array(Masses)

        if isinstance(weights_true, list):
            weights_true = np.array(weights_true)

        Subfaces_matrix = ut.subfaces_list_to_matrix(Subfaces, self.dimension)
        Subfaces_true_matrix = ut.subfaces_list_to_matrix(subfaces_true, self.dimension)

        if not self.include_singletons:
            id_keep_estim = np.where(np.sum(Subfaces_matrix, axis=1) >= 2)[0]
            id_keep_true = np.where(np.sum(Subfaces_true_matrix, axis=1) >= 2)[0]
            Subfaces_matrix = Subfaces_matrix[id_keep_estim]
            Masses = Masses[id_keep_estim]
            Masses = Masses / np.sum(Masses) if np.sum(Masses) > 0 else Masses
            Subfaces_true_matrix = Subfaces_true_matrix[id_keep_true]
            weights_true = weights_true[id_keep_true]
            weights_true = (
                weights_true / np.sum(weights_true)
                if np.sum(weights_true) > 0
                else weights_true
            )

        est_to_truth = 2 * ut.total_deviance_binary_matrices(
            Subfaces_matrix, Masses, Subfaces_true_matrix, weights_true, self.rate
        )
        truth_to_est = 2 * ut.total_deviance_binary_matrices(
            Subfaces_true_matrix, weights_true, Subfaces_matrix, Masses, self.rate
        )

        return est_to_truth, truth_to_est


# class damex:
#     """
#     An unsupervised feature clustering model to learn "which subsets of
#     features of a random vector are likely to be simultaneously large, while
#     the others are small".
#     """

#     def __init__(self, epsilon=0.1, min_counts=0, thresh_train=None,
#                  thresh_test=None,
#                  include_singletons=False,
#                  rate=10):
#         """
#         Initialize the DAMEX model with specified parameters.

#         Parameters:
#         - epsilon (float): Tolerance level for clustering.

#         - min_counts (int): Minimum number of points required to form
#             a cluster.

#         - thresh_train (float): Threshold for training data.

#         - thresh_test (float): Threshold for test data.

#         - include_singletons (bool): Whether to include singletons in
#               training and testing.

#         - rate (float, >0): rate parameter for the dispersion model
#             on subfaces.
#         """
#         self.epsilon = epsilon
#         self.min_counts = min_counts
#         self.thresh_train = thresh_train
#         self.thresh_test = thresh_test
#         self.include_singletons = include_singletons
#         self.rate = rate
#         self.subfaces = None  # Identified subfaces
#         self.masses = None  # Masses associated with subfaces
#         self.total_mass = None  # Total mass of extremes
#         self.number_extremes = None  # Number of extreme points
#         self.dimension = None  # Dimensionality of the data

#     def fit(self, X, epsilon=None, standardize=True):
#         """
#         Fit the DAMEX model to the data.

#         Parameters:
#         ------------

#         - X (np.ndarray): Input data.

#         - threshold (float): Threshold for identifying extremes.

#         - epsilon (float): Tolerance level for clustering.
#             If None, the instance's attribute will be used.

#         - standardize (bool): Whether to standardize the data.


#         Returns:
#         ---------

#         - Subfaces (list): Identified subfaces.

#         - Masses (np.ndarray): Masses associated with subfaces.
#         """
#         # Update instance attributes with optional arguments passed
#         if epsilon is not None:
#             self.epsilon = epsilon

#         # Record dimension
#         self.dimension = np.shape(X)[1]

#         # Standardize X if needed
#         Xt = rank_transform(X) if standardize else X
#         norm_Xt = np.max(Xt, axis=1)

#         # Set the training threshold if not provided
#         if self.thresh_train is None:
#             self.thresh_train = np.quantile(
#                 norm_Xt, 1 - 1 / np.sqrt(len(norm_Xt)))

#         # Update total mass accordingly
#         self.number_extremes = np.sum(norm_Xt > self.thresh_train)
#         self.total_mass = self.thresh_train * self.number_extremes / len(norm_Xt)

#         # Fit the model
#         Subfaces, Masses = ut.damex_fit(
#             Xt, self.thresh_train, self.epsilon, self.min_counts,
#             standardize=False, include_singletons=self.include_singletons)

#         # Update instance's attributes
#         self.subfaces = Subfaces
#         self.masses = Masses
#         return Subfaces, Masses

#     def deviance(self, Xtest, thresh_test=None, standardize=False):
#         """Calculate the deviance of the model on test data.

#         Parameters:
#         -----------

#         - Xtest (np.ndarray): Test data.

#         - thresh_test(float): Threshold for identifying extremes in
#             the test set.  If None, the instance's attribute will be
#             used instead. If the latter is also None, thresh_test
#             will be set to the same value as thresh_train.

#         - standardize (bool): Whether to standardize the data.

#         Returns:
#         --------

#         - float: Deviance value.

#         """
#         if self.subfaces is None:
#             raise RuntimeError("The model has not been fitted yet. Call 'fit' "
#                                "with appropriate arguments before using this "
#                                "method.")

#         Xt = rank_transform(Xtest) if standardize else Xtest

#         # update test threshold if passed as argument
#         if thresh_test is not None:
#             self.thresh_test = thresh_test
#         # set thresh_test to thresh_train if not previously specified
#         if self.thresh_test is None:
#             self.thresh_test = self.thresh_train

#         Subfaces = self.subfaces
#         Masses = self.masses

#         negative_pseudo_lkl = ut.total_deviance(
#             Subfaces, Masses, Xt, self.thresh_test,
#             self.include_singletons, self.epsilon, rate=self.rate)

#         return 2 * negative_pseudo_lkl

#     def get_AIC(self, Xtrain,  standardize=True):
#         """
#         Calculate the Akaike Information Criterion (AIC) for the model.

#         Parameters:
#         -----------

#         - Xtrain (np.ndarray): Training data.

#         - standardize (bool): Whether to standardize the data.

#         Returns:
#         ---------

#         - float: AIC value.
#         """
#         if self.masses is None:
#             raise RuntimeError("Fit the model before computing the AIC")

#         Masses = self.masses
#         intern_deviance = self.deviance(Xtrain, self.thresh_train, standardize)
#         return intern_deviance + 2 * len(Masses) / self.number_extremes

#     def select_epsilon_AIC(self, grid, X, standardize=True,
#                            unstable_eps_max=0.05, plot=False,
#                            update_epsilon=False):
#         """
#         Select the optimal epsilon value based on AIC.

#         Parameters:
#         ----------

#         - grid (list): Grid of epsilon values to test.

#         - X (np.ndarray): Input data.

#         - standardize (bool): Whether to standardize the data.

#         - unstable_eps_max (float): Maximum epsilon value for unstable
#           solutions.

#         - plot (bool): Whether to plot the AIC values.

#         - update_epsilon (bool): Whether to update the model's epsilon value.

#         Returns:
#         --------

#         - eps_select_aic (float): Selected epsilon value.

#         - aic_opt (float): Optimal AIC value.

#         - vect_aic (np.ndarray): Vector of AIC values.
#         """
#         old_epsilon = deepcopy(self.epsilon)
#         Xt = rank_transform(X) if standardize else X
#         ntests = len(grid)
#         vect_aic = np.zeros(ntests)

#         for counter, eps in enumerate(grid):
#             _, _ = self.fit(Xt,  eps, standardize=False)
#             vect_aic[counter] = self.get_AIC(Xt, standardize=False)

#         i_maxerr = np.argmax(vect_aic[grid < unstable_eps_max])
#         eps_maxerr = grid[i_maxerr]
#         i_mask = grid <= eps_maxerr
#         i_minAIC = np.argmin(vect_aic + (1e+23) * i_mask)
#         eps_select_aic = grid[i_minAIC]
#         aic_opt = vect_aic[i_minAIC]

#         if plot:
#             print(f'DAMEX: selected epsilon with AIC criterion '
#                   f'{round_signif(eps_select_aic, 2)}')
#             plt.figure(figsize=(10, 5))
#             plt.xlabel('epsilon')
#             plt.ylabel('AIC')
#             plt.title('DAMEX- AIC versus epsilon')
#             plt.scatter(grid, vect_aic, c='gray', label='AIC')
#             plt.plot([eps_select_aic, eps_select_aic], [0, max(vect_aic)],
#                      c='red')
#             plt.grid(True)
#             plt.show()

#         # Reset instance's epsilon value with initial or selected value
#         if update_epsilon:
#             self.epsilon = eps_select_aic
#         else:
#             self.epsilon = old_epsilon

#         # Re-fit the model with initial or final epsilon
#         _, _ = self.fit(Xt, self.epsilon, False)

#         return eps_select_aic, aic_opt, vect_aic

#     def deviance_CV(self, X, standardize=True, epsilon=None,
#                     thresh_test=None, cv=5, random_state=None):
#         """
#         Estimate the expected deviance using cross-validation.

#         Parameters:
#         -----------

#         - X (np.ndarray): Input data.

#         - standardize (bool): Whether to standardize the data. default to True.

#         - epsilon (float): Tolerance level for clustering. Default to
#           None. If None, the instance's attribute will be used.

#         - thresh_test (float, optional): radial threshold for test
#           sets. If None, the instance's attribute will be used. If the
#           latter is also None, the training threshold is used instead.

#         - cv (int): Number of cross-validation folds. Default to 5

#         - random_state (int): Random seed for reproducibility. Default to None

#         Returns:
#         --------

#         - np.ndarray: Cross-validated deviance scores.

#         """
#         Xt = rank_transform(X) if standardize else X

#         # Update instance's attributes if passed as arguments
#         if epsilon is not None:
#             self.epsilon = epsilon
#         if thresh_test is not None:
#             self.thresh_test = thresh_test
#         if self.thresh_test is None:
#             self.thresh_test = self.thresh_train

#         cv_neglkl_scores = ut.ftclust_cross_validate(
#             Xt, standardize=False, algo='damex', tolerance=self.epsilon,
#             min_counts=self.min_counts, use_max_subfaces=False,
#             thresh_train=self.thresh_train, thresh_test=self.thresh_test,
#             include_singletons_train=self.include_singletons,
#             include_singletons_test=self.include_singletons,
#             rate=self.rate, cv=cv, random_state=random_state)

#         return 2 * cv_neglkl_scores

#     def select_epsilon_CV(self, grid, X, standardize=True,
#                           update_epsilon=False,
#                           unstable_tol_max=0.05, thresh_test=None,  cv=5,
#                           random_state=None, plot=False):
#         """
#         Select the optimal epsilon value based on cross-validation.

#         Parameters:
#         ------------

#         - grid (list): Grid of epsilon values to test.

#         - X (np.ndarray): Input data.

#         - standardize (bool): Whether to standardize the data.

#         - update_epsilon (bool): Whether to update the model's epsilon value.

#         - unstable_tol_max (float): Maximum tolerance value for unstable
#           solutions.

#         - cv (int): Number of cross-validation folds.

#         - random_state (int): Random seed for reproducibility.

#         - plot (bool): Whether to plot the CV deviance values.


#         Returns:

#         - tol_cv (float): Selected epsilon value.

#         - deviance_tol_cv (float): Deviance value for the selected epsilon.

#         - cv_deviance_vect (np.ndarray): Vector of CV deviance values.
#         """
#         Xt = rank_transform(X) if standardize else X
#         epsilon_old = deepcopy(self.epsilon)
#         ntol = len(grid)
#         cv_deviance_vect = np.zeros(ntol)

#         for i, eps in enumerate(grid):
#             cv_scores = self.deviance_CV(Xt, False, eps, thresh_test,
#                                          cv, random_state)
#             cv_deviance_vect[i] = np.mean(cv_scores)

#         i_maxerr = np.argmax(cv_deviance_vect[grid < unstable_tol_max])
#         tol_maxerr = grid[i_maxerr]
#         i_mask = grid <= tol_maxerr
#         i_cv = np.argmin(cv_deviance_vect + (1e+23) * i_mask)
#         tol_cv = grid[i_cv]
#         deviance_tol_cv = cv_deviance_vect[i_cv]

#         if plot:
#             print(f'DAMEX: selected epsilon with CV criterion '
#                   f'{round_signif(tol_cv, 2)}')
#             plt.scatter(grid, cv_deviance_vect, c='gray', label='CV deviance')
#             plt.plot([tol_cv, tol_cv], [0, deviance_tol_cv], c='red',
#                      label='selected value')
#             plt.grid()
#             plt.legend()
#             plt.title("DAMEX: clustering pseudo-deviance, K-fold CV scores")
#             plt.show()

#         if update_epsilon:
#             self.epsilon = tol_cv
#         else:
#             self.epsilon = epsilon_old

#         # Re-fit the model with final epsilon
#         _, _ = self.fit(Xt, False)

#         return tol_cv, deviance_tol_cv, cv_deviance_vect

#     def deviance_to_true(self, subfaces_true, weights_true):
#         """
#         Calculate the deviance between the estimated subfaces and the true
#         subfaces.

#         Parameters:
#         -----------
#         - subfaces_true (list): True subfaces.

#         - weights_true (list): Weights of the true subfaces.

#         Returns:
#         --------

#         - est_to_truth (float): Deviance from estimated to true subfaces.

#         - truth_to_est (float): Deviance from true to estimated subfaces.
#         """
#         if self.subfaces is None:
#             raise RuntimeError("DAMEX has not been fitted yet")
#         Subfaces = self.subfaces
#         Masses = self.masses

#         if isinstance(Masses, list):
#             Masses = np.array(Masses)

#         if isinstance(weights_true, list):
#             weights_true = np.array(weights_true)

#         Subfaces_matrix = ut.subfaces_list_to_matrix(
#             Subfaces, self.dimension)
#         Subfaces_true_matrix = ut.subfaces_list_to_matrix(
#             subfaces_true, self.dimension)

#         if not self.include_singletons:
#             id_keep_estim = np.where(np.sum(Subfaces_matrix, axis=1) >= 2)[0]
#             id_keep_true = np.where(np.sum(Subfaces_true_matrix,
#                                            axis=1) >= 2)[0]
#             Subfaces_matrix = Subfaces_matrix[id_keep_estim]
#             Masses = Masses[id_keep_estim]
#             Masses = Masses / np.sum(Masses) if np.sum(Masses) > 0 else Masses
#             Subfaces_true_matrix = Subfaces_true_matrix[id_keep_true]
#             weights_true = weights_true[id_keep_true]
#             weights_true = weights_true / np.sum(weights_true) \
#                 if np.sum(weights_true) > 0 else weights_true

#         est_to_truth = 2 * ut.total_deviance_binary_matrices(
#             Subfaces_matrix, Masses, Subfaces_true_matrix, weights_true,
#             self.rate)
#         truth_to_est = 2 * ut.total_deviance_binary_matrices(
#             Subfaces_true_matrix, weights_true, Subfaces_matrix, Masses,
#             self.rate)

#         return est_to_truth, truth_to_est



import numpy as np
import matplotlib.pyplot as plt
from copy import deepcopy
from . import utilities as ut  # #import binary_large_features
from MLExtreme.utils.EVT_basics import rank_transform, round_signif


class clef:
    """
    A class for performing clustering of large extreme features (CLEF) on a dataset.

    This class provides methods to fit the CLEF model to data, calculate deviance, compute the
    Akaike Information Criterion (AIC), select optimal kappa_min values based on AIC or
    cross-validation, and calculate the deviance between estimated and true subfaces.

    Parameters:
    -----------
    kappa_min : float, default=0.1
        Tolerance level for clustering.
    thresh_train : float, optional
        Threshold for training data.
    thresh_test : float, optional
        Threshold for test data.
    rate : float, default=10
        Rate parameter for the dispersion model on subfaces.

    Attributes:
    -----------
    subfaces : list
        Identified subfaces.
    masses : array-like
        Masses associated with subfaces.
    total_mass : float
        Total mass of extremes.
    number_extremes : int
        Number of extreme points.
    dimension : int
        Dimensionality of the data.
    """

    def __init__(self, kappa_min=0.1, thresh_train=None, thresh_test=None, rate=10):
        """
        Initialize the CLEF model with specified parameters.

        Parameters:
        -----------
        kappa_min : float, default=0.1
            Tolerance level for clustering.
        thresh_train : float, optional
            Threshold for training data.
        thresh_test : float, optional
            Threshold for test data.
        rate : float, default=10
            Rate parameter for the dispersion model on subfaces.
        """
        self.kappa_min = kappa_min
        self.thresh_train = thresh_train
        self.thresh_test = thresh_test
        self.rate = rate
        self.subfaces = None  # Identified subfaces
        self.masses = None  # Masses associated with subfaces
        self.total_mass = None  # Total mass of extremes
        self.number_extremes = None  # Number of extreme points
        self.dimension = None  # Dimensionality of the data

    def fit(self, X, kappa_min=None, standardize=True):
        """
        Fit the CLEF model to the data.

        Parameters:
        -----------
        X : np.ndarray
            Input data.
        kappa_min : float, optional
            Tolerance level for clustering. If None, the instance's attribute will be used.
        standardize : bool, default=True
            Whether to standardize the data.

        Returns:
        --------
        Subfaces : list
            Identified subfaces.
        Masses : np.ndarray
            Masses associated with subfaces.
        """
        # Update instance attributes with optional arguments passed
        if kappa_min is not None:
            self.kappa_min = kappa_min

        # Record dimension
        self.dimension = np.shape(X)[1]

        # Standardize X if needed
        Xt = rank_transform(X) if standardize else X
        norm_Xt = np.max(Xt, axis=1)

        # Set the training threshold if not provided
        if self.thresh_train is None:
            self.thresh_train = np.quantile(norm_Xt, 1 - 1 / np.sqrt(len(norm_Xt)))

        # Update total mass accordingly
        self.number_extremes = np.sum(norm_Xt > self.thresh_train)
        self.total_mass = self.thresh_train * self.number_extremes / len(norm_Xt)

        # Fit the model
        Subfaces = ut.clef_fit(
            Xt,
            self.thresh_train,
            self.kappa_min,
            standardize=False,
            include_singletons=False,
        )
        Masses = ut.estim_subfaces_mass(
            Subfaces, Xt, self.thresh_train, epsilon=None, standardize=False
        )
        # Update instance's attributes
        self.subfaces = Subfaces
        self.masses = Masses

        return Subfaces, Masses

    def deviance(self, Xtest, thresh_test=None, standardize=False):
        """Calculate the deviance of the (trained) model on test data, normalized by the extreme sample size.

        First the test data is transformed into a binary matrix, say
        `testBinary`. Each row corresponds to a sample point which
        norm exceeds `thresh_test`.  The binary entries indicates
        which coordinates of a given extreme point exceed the quantity
        `epsilon * thresh_test`.

        Then the subfaces previously learnt by the model (stored in
        `self.subfaces`), together with the associated masses, are
        compared to `testBinary` in a a dispersion model.

        The basic building block of the dispersion model is a unit
        deviance d(subface_i, subface_j) defined between subfaces, as
        the ratio between the cardinalities of the symmetric
        difference and the union of the two sets of coordinates
        defining the suface, see
        :func:`unsupervised.feature_clustering.utilities.unit_deviance`. This is the same unit deviance as the one proposed in Reference [4]  below 

        The total (normalized) deviance is then defined as
        
        .. math::

            \\frac{- 2}{n_{extremes}}\\sum_{i=1}^{n_{extremes}} \
            \\log\\Big( \\sum_{j=1}^{n_{faces}} \
            w_j \\exp(- rate *d(testBinary_i, subface_j) ) \\Big), 


        where `n_{extremes}` is the number of samples which radius
        exceeds `thresh_test` in `Xtest`, rate=`self.rate`. It may be seen as

        "Twice the negative pseudo-likelihood in a dispersion model à la Jorgensen, divided by the extreme sample size"

        References:
        ------------
        [1] Cordeiro, G. M., Labouriau, R., & Botter, D. A. (2021). \
        An introduction to Bent Jørgensen’s ideas. Brazilian Journal of Probability and Statistics, 35(1), 2-20.

        [2] Jørgensen, B. (1987). Exponential dispersion models. \
        Journal of the Royal Statistical Society Series B: Statistical Methodology, \
        49(2), 127-145.

        [3] Jørgensen, B. (1997). The theory of dispersion models. CRC Press.

        [4] Chiapino, M., & Sabourin, A. (2016, September). Feature clustering for extreme events analysis, with application to extreme stream-flow data. In International workshop on new frontiers in mining complex patterns (pp. 132-147). Cham: Springer International Publishing.

        Parameters:
        -----------
        Xtest : np.ndarray
            Test data.
        thresh_test : float, optional
            Threshold for identifying extremes in the test set.
            If None, the instance's attribute will be used.
        standardize : bool, default=False
            Whether to standardize the data.

        Returns:
        --------
        float
            Deviance value,that is twice the negative pseudo-likelihood, divided by the number of extremes.

        """
        if self.subfaces is None:
            raise RuntimeError(
                "The model has not been fitted yet. Call 'fit' "
                "with appropriate arguments before using this "
                "method."
            )

        Xt = rank_transform(Xtest) if standardize else Xtest

        # Update instance's attributes if passed as arguments
        if thresh_test is not None:
            self.thresh_test = thresh_test
        if self.thresh_test is None:
            self.thresh_test = self.thresh_train

        Subfaces = self.subfaces
        Masses = self.masses

        negative_pseudo_lkl = ut.total_deviance(
            Subfaces, Masses, Xt, self.thresh_test, False, None,
            rate=self.rate
        )

        return 2 * negative_pseudo_lkl

    def get_AIC(self, Xtrain, standardize=True):
        """Calculate the Akaike Information Criterion (AIC) for the
        model incorporating the normalized deviance from
        :func:`self.deviance` and the (normalized again) number of
        components in the model, that is the number of subfaces with
        positive extreme mass. More precisely the function returns

        self.deviance  + 2 * number_of_subfaces/number_of_extremes

        where whe recall that self.deviance is:

        "twice the negative pseudo log-likelihood in a dispersion model"

        

        Parameters:
        -----------
        Xtrain : np.ndarray
            Training data.
        standardize : bool, default=True
            Whether to standardize the data.

        Returns:
        --------
        float
            AIC value.

        """
        if self.masses is None:
            raise RuntimeError("Fit the model before computing the AIC")

        Masses = self.masses
        intern_deviance = self.deviance(Xtrain, self.thresh_train, standardize)
        return intern_deviance + 2 * len(Masses) / self.number_extremes

    def select_kappa_min_AIC(
        self,
        grid,
        X,
        standardize=True,
        unstable_kappam_max=0.05,
        plot=False,
        update_kappa_min=False,
    ):
        """
        Select the optimal kappa_min value based on AIC.

        Parameters:
        -----------
        grid : list
            Grid of kappa_min values to test.
        X : np.ndarray
            Input data.
        standardize : bool, default=True
            Whether to standardize the data.
        unstable_kappam_max : float, default=0.05
            Maximum kappa_min value for unstable solutions. The selected value returned by the function will be greater than this parameter. 
        plot : bool, default=False
            Whether to plot the AIC values.
        update_kappa_min : bool, default=False
            Whether to update the model's kappa_min value.

        Returns:
        --------
        kappam_select_aic : float
            Selected kappa_min value.
        aic_opt : float
            Optimal AIC value.
        vect_aic : np.ndarray
            Vector of AIC values.
        """
        old_kappa_min = deepcopy(self.kappa_min)
        Xt = rank_transform(X) if standardize else X
        ntests = len(grid)
        vect_aic = np.zeros(ntests)

        for counter, kapp in enumerate(grid):
            _, _ = self.fit(Xt, kapp, standardize=False)
            vect_aic[counter] = self.get_AIC(Xt, standardize=False)

        i_maxerr = np.argmax(vect_aic[grid < unstable_kappam_max])
        kapp_maxerr = grid[i_maxerr]
        i_mask = grid <= kapp_maxerr
        i_minAIC = np.argmin(vect_aic + (1e23) * i_mask)
        kapp_select_aic = grid[i_minAIC]
        aic_opt = vect_aic[i_minAIC]

        if plot:
            print(
                f"CLEF: selected kappa_min with AIC criterion: "
                f"{round_signif(kapp_select_aic, 2)}"
            )
            plt.figure(figsize=(10, 5))
            plt.xlabel("kappa_min")
            plt.ylabel("AIC")
            plt.title("CLEF: AIC versus kappa_min")
            plt.scatter(grid, vect_aic, c="gray", label="AIC")
            plt.plot([kapp_select_aic, kapp_select_aic], [0, aic_opt], c="red")
            plt.grid(True)
            plt.show()

        # Update instance's kappa_min value with initial or selected value
        if update_kappa_min:
            self.kappa_min = kapp_select_aic
        else:
            self.kappa_min = old_kappa_min

        # Re-fit the model with final kappa_min
        _, _ = self.fit(Xt, self.kappa_min, False)

        return kapp_select_aic, aic_opt, vect_aic

    def deviance_CV(
        self,
        X,
        standardize=True,
        kappa_min=None,
        thresh_test=None,
        cv=5,
        random_state=None,
    ):
        """
        Estimate the expected deviance using cross-validation. See :func:`deviance`

        Parameters:
        -----------
        X : np.ndarray
            Input data.
        standardize : bool, default=True
            Whether to standardize the data.
        kappa_min : float, optional
            Tolerance level for clustering. If None, the instance's attribute will be used.
        thresh_test : float, optional
            Radial threshold for test sets. If None, the instance's attribute will be used.
        cv : int, default=5
            Number of cross-validation folds.
        random_state : int, optional
            Random seed for reproducibility.

        Returns:
        --------
        np.ndarray
            Cross-validated deviance scores.
        """
        Xt = rank_transform(X) if standardize else X

        # Update instance's attributes if passed as arguments
        if kappa_min is not None:
            self.kappa_min = kappa_min
        if thresh_test is not None:
            self.thresh_test = thresh_test
        if self.thresh_test is None:
            self.thresh_test = self.thresh_train

        cv_neglkl_scores = ut.ftclust_cross_validate(
            Xt,
            standardize=False,
            algo="clef",
            tolerance=self.kappa_min,
            min_counts=None,
            use_max_subfaces=None,
            thresh_train=self.thresh_train,
            thresh_test=self.thresh_test,
            include_singletons_train=False,
            include_singletons_test=False,
            rate=self.rate,
            cv=cv,
            random_state=random_state,
        )

        return 2 * cv_neglkl_scores

    def select_kappa_min_CV(
        self,
        grid,
        X,
        standardize=True,
        update_kappa_min=False,
        unstable_tol_max=0.05,
        thresh_test=None,
        cv=5,
        random_state=None,
        plot=False,
    ):
        """
        Select the optimal kappa_min value based on cross-validation.

        Parameters:
        -----------
        grid : list
            Grid of kappa_min values to test.
        X : np.ndarray
            Input data.
        standardize : bool, default=True
            Whether to standardize the data.
        update_kappa_min : bool, default=False
            Whether to update the model's kappa_min value.
        unstable_tol_max : float, default=0.05
            Maximum tolerance value for unstable solutions.
        cv : int, default=5
            Number of cross-validation folds.
        random_state : int, optional
            Random seed for reproducibility.
        plot : bool, default=False
            Whether to plot the CV deviance values.

        Returns:
        --------
        tol_cv : float
            Selected kappa_min value.
        deviance_tol_cv : float
            Deviance value for the selected kappa_min.
        cv_deviance_vect : np.ndarray
            Vector of CV deviance values.
        """
        Xt = rank_transform(X) if standardize else X
        kappa_min_old = deepcopy(self.kappa_min)
        ntol = len(grid)
        cv_deviance_vect = np.zeros(ntol)

        for i, kapp in enumerate(grid):
            cv_scores = self.deviance_CV(Xt, False, kapp, thresh_test, cv, random_state)
            cv_deviance_vect[i] = np.mean(cv_scores)

        i_maxerr = np.argmax(cv_deviance_vect[grid < unstable_tol_max])
        tol_maxerr = grid[i_maxerr]
        i_mask = grid <= tol_maxerr
        i_cv = np.argmin(cv_deviance_vect + (1e23) * i_mask)
        tol_cv = grid[i_cv]
        deviance_tol_cv = cv_deviance_vect[i_cv]

        if plot:
            print(
                f"CLEF: selected kappa_min with CV estimate of deviance: "
                f"{round_signif(tol_cv, 2)}"
            )
            plt.scatter(grid, cv_deviance_vect, c="gray", label="CV deviance")
            plt.plot(
                [tol_cv, tol_cv], [0, deviance_tol_cv], c="red",
                label="selected value"
            )
            plt.grid()
            plt.legend()
            plt.title("CLEF: expected deviance estimated with K-fold CV")
            plt.show()

        if update_kappa_min:
            self.kappa_min = tol_cv
        else:
            self.kappa_min = kappa_min_old

        # Re-fit the model with final kappa_min
        _, _ = self.fit(Xt, self.kappa_min, False)

        return tol_cv, deviance_tol_cv, cv_deviance_vect

    def deviance_to_true(self, subfaces_true, weights_true):
        """Calculate the deviance between the estimated subfaces and
        the true subfaces.

        The function  returns a tuple (est_to_truth, truth_to_est) respectively
        defined as follows:

        .. math::

            \\text{est-to-truth} = -2\
        \\sum_{j}^{\\text{N-estimated}} \
        w^{estim}_j \\log \\Big( \
        \\sum_{i=1}^{\\text{N-true}}\
       w^{true}_i e^{-\\text{rate}*d(\\text{estimated-subface}_j, \
        \\text{true-subface}_i)} \
        \\Big)
        
            \\text{tuth-to-est} = -2\
        \\sum_{i=1}^{\\text{N-true}}\
         w^{true}_i \\log \\Big( \
        \\sum_{j}^{\\text{N-estimated}} \
        w^{estim}_j 
         e^{-\\text{rate}*d(\\text{estimated-subface}_j, \
        \\text{true-subface}_i)} \
        \\Big)
        
        where :math:`w^{estim}_j =` self.mass[j-1] / sum(self.masses) and
        :math:`w^{truth}_j` is the mass of the considered true subface divided
        by the total mass of extremes 
        
        Parameters:
        -----------
        subfaces_true : list
            True subfaces.
        weights_true : list
            Weights of the true subfaces.

        Returns:
        --------
        est_to_truth : float
            Deviance from estimated to true subfaces.
        truth_to_est : float
            Deviance from true to estimated subfaces.

        """
        if self.subfaces is None:
            raise RuntimeError("CLEF has not been fitted yet")

        Subfaces_matrix = ut.subfaces_list_to_matrix(self.subfaces, self.dimension)
        Masses = self.masses
        if isinstance(self.masses, list):
            Masses = np.array(Masses)

        Subfaces_true_matrix = ut.subfaces_list_to_matrix(subfaces_true, self.dimension)
        if isinstance(weights_true, list):
            weights_true = np.array(weights_true)

        # if not self.include_singletons:
        id_keep_estim = np.where(np.sum(Subfaces_matrix, axis=1) >= 2)[0]
        id_keep_true = np.where(np.sum(Subfaces_true_matrix, axis=1) >= 2)[0]
        Subfaces_matrix = Subfaces_matrix[id_keep_estim]
        Masses = Masses[id_keep_estim]
        Masses = Masses / np.sum(Masses) if np.sum(Masses) > 0 else Masses
        Subfaces_true_matrix = Subfaces_true_matrix[id_keep_true]
        weights_true = weights_true[id_keep_true]
        weights_true = (
            weights_true / np.sum(weights_true)
            if np.sum(weights_true) > 0
            else weights_true
        )

        est_to_truth = 2 * ut.total_deviance_binary_matrices(
            Subfaces_matrix, Masses, Subfaces_true_matrix, weights_true, self.rate
        )
        truth_to_est = 2 * ut.total_deviance_binary_matrices(
            Subfaces_true_matrix, weights_true, Subfaces_matrix, Masses, self.rate
        )

        return est_to_truth, truth_to_est


# class clef:
#     def __init__(self, kappa_min=0.1, thresh_train=None,
#                  thresh_test=None,
#                  rate=10):
#         """Initialize the DAMEX model with specified parameters.

#         Parameters:
#         -----------

#         - kappa_min (float): Tolerance level for clustering.

#         - min_counts (int): Minimum number of points required to
#             form a cluster.

#         - thresh_train (float): Threshold for training data.

#         - thresh_test (float): Threshold for test data.

#          """
#         self.kappa_min = kappa_min
#         self.thresh_train = thresh_train
#         self.thresh_test = thresh_test
#         self.rate = rate
#         self.subfaces = None  # Identified subfaces
#         self.masses = None  # Masses associated with subfaces
#         self.total_mass = None  # Total mass of extremes
#         self.number_extremes = None  # Number of extreme points
#         self.dimension = None  # Dimensionality of the data

#     def fit(self, X,  kappa_min=None, standardize=True):
#         """Fit the CLEF model to the data.

#         Parameters:
#         -----------

#         - X (np.ndarray): Input data.

#         - kappa_min
#         (float): Tolerance level for clustering. If None, the
#         instance's attribute will be used.

#         - standardize (bool):
#         Whether to standardize the data.


#         Returns:
#         ---------

#         - Subfaces (list): Identified subfaces.

#         - Masses (np.ndarray): Masses associated with subfaces.

#         """
#         # Update instance attributes with optional arguments passed
#         if kappa_min is not None:
#             self.kappa_min = kappa_min

#         # Record dimension
#         self.dimension = np.shape(X)[1]

#         # Standardize X if needed
#         Xt = rank_transform(X) if standardize else X
#         norm_Xt = np.max(Xt, axis=1)

#         # Set the training threshold if not provided
#         if self.thresh_train is None:
#             self.thresh_train = np.quantile(
#                 norm_Xt, 1 - 1 / np.sqrt(len(norm_Xt)))

#         # Update total mass accordingly
#         self.number_extremes = np.sum(norm_Xt > self.thresh_train)
#         self.total_mass = self.thresh_train * \
#             self.number_extremes / len(norm_Xt)

#         # Fit the model
#         Subfaces = ut.clef_fit(
#             Xt, self.thresh_train, self.kappa_min,
#             standardize=False, include_singletons=False)
#         Masses = ut.estim_subfaces_mass(
#             Subfaces, Xt, self.thresh_train, epsilon=None,
#             standardize=False)
#         # Update instance's attributes
#         self.subfaces = Subfaces
#         self.masses = Masses

#         return Subfaces, Masses

#     def deviance(self, Xtest, thresh_test=None,  standardize=False):
#         """
#         Calculate the deviance of the model on test data.

#         Parameters:
#         -----------

#         - Xtest (np.ndarray): Test data.

#         - thresh_test
#         (float): Threshold for identifying extremes in the test set.
#         If None, the instance's attribute will be used instead. If the
#         lmatter is also None, thresh_test will be set to the same
#         value as thresh_train.

#         - standardize (bool): Whether to standardize the data.

#         Returns:
#         ---------

#         - float: Deviance value.
#         """
#         if self.subfaces is None:
#             raise RuntimeError("The model has not been fitted yet. Call 'fit' "
#                                "with appropriate arguments before using this "
#                                "method.")

#         Xt = rank_transform(Xtest) if standardize else Xtest

#         # Update instance's attributes if passed as arguments
#         if thresh_test is not None:
#             self.thresh_test = thresh_test
#         if self.thresh_test is None:
#             self.thresh_test = self.thresh_train

#         Subfaces = self.subfaces
#         Masses = self.masses

#         negative_pseudo_lkl = ut.total_deviance(
#             Subfaces, Masses, Xt,  self.thresh_test,
#             False, None, rate=self.rate)

#         return 2 * negative_pseudo_lkl

#     def get_AIC(self, Xtrain, standardize=True):
#         """
#         Calculate the Akaike Information Criterion (AIC) for the model.

#         Parameters:

#         - Xtrain (np.ndarray): Training data.

#         - standardize (bool): Whether to standardize the data.

#         Returns:

#         - float: AIC value.
#         """
#         if self.masses is None:
#             raise RuntimeError("Fit the model before computing the AIC")

#         Masses = self.masses
#         intern_deviance = self.deviance(Xtrain, self.thresh_train, standardize)
#         return intern_deviance + 2 * len(Masses) / self.number_extremes

#     def select_kappa_min_AIC(self, grid, X, standardize=True,
#                              unstable_kappam_max=0.05,
#                              plot=False,
#                              update_kappa_min=False):
#         """
#         Select the optimal kappa_min value based on AIC.

#         Parameters:
#         ----------

#         - grid (list): Grid of kappa_min values to test.

#         - X (np.ndarray): Input data.

#         - standardize (bool): Whether to standardize the data.

#         - unstable_kappam_max (float): Maximum kappa_min value for unstable
#           solutions.

#         - plot (bool): Whether to plot the AIC values.

#         - update_kappa_min (bool): Whether to update the model's
#             kappa_min value.

#         Returns:
#         --------

#         - kappam_select_aic (float): Selected kappa_min value.

#         - aic_opt (float): Optimal AIC value.

#         - vect_aic (np.ndarray): Vector of AIC values.
#         """
#         old_kappa_min = deepcopy(self.kappa_min)
#         Xt = rank_transform(X) if standardize else X
#         ntests = len(grid)
#         vect_aic = np.zeros(ntests)

#         for counter, kapp in enumerate(grid):
#             _, _ = self.fit(Xt, kapp, standardize=False)
#             vect_aic[counter] = self.get_AIC(Xt, standardize=False)

#         i_maxerr = np.argmax(vect_aic[grid < unstable_kappam_max])
#         kapp_maxerr = grid[i_maxerr]
#         i_mask = grid <= kapp_maxerr
#         i_minAIC = np.argmin(vect_aic + (1e+23) * i_mask)
#         kapp_select_aic = grid[i_minAIC]
#         aic_opt = vect_aic[i_minAIC]

#         if plot:
#             print(f'CLEF: selected kappa_min with AIC criterion: '
#                   f'{round_signif(kapp_select_aic, 2)}')
#             plt.figure(figsize=(10, 5))
#             plt.xlabel('kappa_min')
#             plt.ylabel('AIC')
#             plt.title('CLEF: AIC versus kappa_min')
#             plt.scatter(grid, vect_aic, c='gray', label='AIC')
#             plt.plot([kapp_select_aic, kapp_select_aic], [0, aic_opt],
#                      c='red')
#             plt.grid(True)
#             plt.show()

#         # Update instance's kappa_min value with initial or selected value
#         if update_kappa_min:
#             self.kappa_min = kapp_select_aic
#         else:
#             self.kappa_min = old_kappa_min

#         # Re-fit the model with final kappa_min
#         _, _ = self.fit(Xt,  self.kappa_min, False)

#         return kapp_select_aic, aic_opt, vect_aic

#     def deviance_CV(self, X, standardize=True, kappa_min=None,
#                     thresh_test=None, cv=5, random_state=None):
#         """
#         Estimate the expected  deviance using cross-validation.

#         Parameters:
#         -----------

#         - X (np.ndarray): Input data.

#         - standardize (bool): Whether to standardize the data. Default to True

#         - kappa_min (float): Tolerance level for clustering. Default to
#           None. If None, the instance's attribute will be used.

#         - thresh_test (float, optional): radial threshold for test
#           sets. If None, the instance's attribute will be used. If the
#           latter is also None, the training threshold is used instead.

#         - cv (int): Number of cross-validation folds. Default to 5

#         - random_state (int): Random seed for reproducibility. Default to None

#         Returns:
#         - np.ndarray: Cross-validated deviance scores.
#         """
#         Xt = rank_transform(X) if standardize else X

#         # Update instance's attributes if passed as arguments
#         if kappa_min is not None:
#             self.kappa_min = kappa_min
#         if thresh_test is not None:
#             self.thresh_test = thresh_test
#         if self.thresh_test is None:
#             self.thresh_test = self.thresh_train

#         cv_neglkl_scores = ut.ftclust_cross_validate(
#             Xt, standardize=False, algo='clef', tolerance=self.kappa_min,
#             min_counts=None, use_max_subfaces=None,
#             thresh_train=self.thresh_train, thresh_test=self.thresh_test,
#             include_singletons_train=False,
#             include_singletons_test=False,
#             rate=self.rate, cv=cv, random_state=random_state)

#         return 2 * cv_neglkl_scores

#     def select_kappa_min_CV(self, grid, X, standardize=True,
#                             update_kappa_min=False,
#                             unstable_tol_max=0.05, thresh_test=None,  cv=5,
#                             random_state=None, plot=False):
#         """
#         Select the optimal kappa_min value based on cross-validation.

#         Parameters:
#         ------------

#         - grid (list): Grid of kappa_min values to test.

#         - X (np.ndarray): Input data.

#         - standardize (bool): Whether to standardize the data.

#         - update_kappa_min (bool): Whether to update the model's
#             kappa_min value.

#         - unstable_tol_max (float): Maximum tolerance value for unstable
#           solutions.

#         - cv (int): Number of cross-validation folds.

#         - random_state (int): Random seed for reproducibility.

#         - plot (bool): Whether to plot the CV deviance values.


#         Returns:

#         - tol_cv (float): Selected kappa_min value.

#         - deviance_tol_cv (float): Deviance value for the selected kappa_min.

#         - cv_deviance_vect (np.ndarray): Vector of CV deviance values.
#         """
#         Xt = rank_transform(X) if standardize else X
#         kappa_min_old = deepcopy(self.kappa_min)
#         ntol = len(grid)
#         cv_deviance_vect = np.zeros(ntol)

#         for i, kapp in enumerate(grid):
#             cv_scores = self.deviance_CV(Xt, False, kapp, thresh_test,
#                                          cv, random_state)
#             cv_deviance_vect[i] = np.mean(cv_scores)

#         i_maxerr = np.argmax(cv_deviance_vect[grid < unstable_tol_max])
#         tol_maxerr = grid[i_maxerr]
#         i_mask = grid <= tol_maxerr
#         i_cv = np.argmin(cv_deviance_vect + (1e+23) * i_mask)
#         tol_cv = grid[i_cv]
#         deviance_tol_cv = cv_deviance_vect[i_cv]

#         if plot:
#             print(f'CLEF: selected kappa_min with CV estimate of deviance: '
#                   f'{round_signif(tol_cv, 2)}')
#             plt.scatter(grid, cv_deviance_vect, c='gray', label='CV deviance')
#             plt.plot([tol_cv, tol_cv], [0, deviance_tol_cv], c='red',
#                      label='selected value')
#             plt.grid()
#             plt.legend()
#             plt.title("CLEF: expected  deviance estimated with  K-fold CV")
#             plt.show()

#         if update_kappa_min:
#             self.kappa_min = tol_cv
#         else:
#             self.kappa_min = kappa_min_old

#         # Re-fit the model with final kappa_min
#         _, _ = self.fit(Xt, self.kappa_min, False)

#         return tol_cv, deviance_tol_cv, cv_deviance_vect

#     def deviance_to_true(self,  subfaces_true, weights_true):
#         """
#         Calculate the deviance between the estimated subfaces and the true
#         subfaces.

#         Parameters:
#         -----------

#         - subfaces_true (list): True subfaces.

#         - weights_true (list): Weights of the true subfaces.

#         - use_max_subfaces (bool): Whether to use maximal subfaces.

#         - rate (float, >0): Rate parameter for deviance calculation.

#         Returns:
#         --------

#         - est_to_truth (float): Deviance from estimated to true subfaces.

#         - truth_to_est (float): Deviance from true to estimated subfaces.
#         """
#         if self.subfaces is None:
#             raise RuntimeError("CLEF has not been fitted yet")

#         Subfaces_matrix = ut.subfaces_list_to_matrix(
#             self.subfaces, self.dimension)
#         Masses = self.masses
#         if isinstance(self.masses, list):
#             Masses = np.array(Masses)

#         Subfaces_true_matrix = ut.subfaces_list_to_matrix(
#             subfaces_true, self.dimension)
#         if isinstance(weights_true, list):
#             weights_true = np.array(weights_true)

#         #if not self.include_singletons:
#         id_keep_estim = np.where(np.sum(Subfaces_matrix, axis=1) >= 2)[0]
#         id_keep_true = np.where(np.sum(Subfaces_true_matrix,
#                                        axis=1) >= 2)[0]
#         Subfaces_matrix = Subfaces_matrix[id_keep_estim]
#         Masses = Masses[id_keep_estim]
#         Masses = Masses / np.sum(Masses) if np.sum(Masses) > 0 else Masses
#         Subfaces_true_matrix = Subfaces_true_matrix[id_keep_true]
#         weights_true = weights_true[id_keep_true]
#         weights_true = weights_true / np.sum(weights_true) \
#             if np.sum(weights_true) > 0 else weights_true

#         est_to_truth = 2 * ut.total_deviance_binary_matrices(
#             Subfaces_matrix, Masses, Subfaces_true_matrix, weights_true,
#             self.rate)
#         truth_to_est = 2 * ut.total_deviance_binary_matrices(
#             Subfaces_true_matrix, weights_true, Subfaces_matrix,
#             Masses,  self.rate)

#         return est_to_truth, truth_to_est



# Author: Anne Sabourin
# Date: 2025
"""
Submodule for Dataset Generation and Plotting Utilities

This submodule provides essential functions for generating datasets and plotting utilities used throughout the MLExtreme package. It focuses on the Pareto Margins convention for modeling multivariate extremes, featuring an angular component that follows a flexible Dirichlet Mixture distribution. Additionally, the submodule includes implementations of the multivariate symmetric Logistic model.

For supervised learning scenarios, the submodule supports the generation of datasets consisting of covariate-target pairs, where the target can be either binary or continuous. The covariates are designed to be regularly varying, and the submodule also includes functionality for rescaling targets.

"""

from scipy.special import loggamma  # gamma, betaln
import numpy as np
import random
import matplotlib.pyplot as plt


def normalize_param_dirimix(Mu, wei):
    """
    Modifies the Dirichlet centers matrix `Mu` and the weights `wei`
    for the Dirichlet mixture model in Pareto margins in order to
    satisfy a moments constraint on the angular distribution of
    extremes.

    The constraint is that :math:`E(\\Theta) = (1/d, \\ldots, 1/d)`, where
    :math:`d` is the dimension of the ambient space. For a Dirichlet
    mixture density with parameters :math:`w = wei, M=Mu`, this is equivalent to

    .. math::

        \\sum_{j=1}^k w_j M_{j, \\cdot} = \\left(\\frac{1}{d}, \\ldots, \\frac{1}{d}\\right)

    The function accepts a matrix `Mu` of dimension (k, d) and a weights
    vector `wei` of dimension (k,). It normalizes these inputs to satisfy
    the moment constraint on the angular distribution, which is derived from
    Pareto-margins standardization. This normalization ensures that each row
    of the resulting `Mu_modif` matrix sums to 1, and the barycenter of the
    rows, weighted by `wei_modif`, is :math:`(1/d, \\ldots, 1/d)`. For further
    details, refer to Boldi & Davison and Sabourin & Naveau. The normalization
    process follows the method described in Chiapino et al. (see references below).

    Parameters
    ----------
    Mu : np.ndarray
        A k x p matrix with positive entries, with rows summing to one.
    wei : np.ndarray
        A weights vector of length k, with nonnegative entries, summing to one.

    Returns
    -------
    tuple
        A tuple containing:
            - Mu1 (np.ndarray): The normalized matrix where columns sum to 1.
            - wei1 (np.ndarray): The updated weights vector of length p.

    Example
    -------
    >>> import numpy as np
    >>> p = 3; k = 2
    >>> Mu0 = 3*np.random.random(2*3).reshape(k, p)
    >>> wei0 = 4*np.random.random(k)
    >>> Mu, wei = normalize_param_dirimix(Mu0, wei0)
    >>> print(f'row sums of Mu : {np.sum(Mu, axis=1)}')
    >>> print(f'barycenter of `Mu` rows with weights `wei` : {np.sum(Mu*wei.reshape(-1, 1), axis=0)}')

    References
    ----------
    [1] Boldi, M. O., & Davison, A. C. (2007). A mixture model for multivariate extremes. Journal of the Royal Statistical Society Series B: Statistical Methodology, 69(2), 217-229.

    [2] Sabourin, A., & Naveau, P. (2014). Bayesian Dirichlet mixture model for multivariate extremes: a re-parametrization. Computational Statistics & Data Analysis, 71, 542-567.

    [3] Chiapino, M., Clémençon, S., Feuillard, V., & Sabourin, A. (2020). A multivariate extreme value theory approach to anomaly clustering and visualization. Computational Statistics, 35(2), 607-628.
    """
    Rho = np.diag(wei) @ Mu
    p = Rho.shape[1]
    Rho_csum = np.sum(Rho, axis=0)  # length p. Barycenter of the input.
    if any(x == 0 for x in Rho_csum):
        raise ValueError("One column of Mu is zero")
    Rho1 = Rho / (Rho_csum * p)  # Rho1 has columns summing to 1/p
    wei1 = np.sum(Rho1, axis=1)  # Updated weights vector of length p
    Mu1 = Rho1 / wei1.reshape(-1, 1)

    return Mu1, wei1


def gen_dirichlet(a, size=1):
    """
    Generate angular random samples from a `Dirichlet distribution <https://en.wikipedia.org/wiki/Dirichlet_distribution>`_ with parameter `a`.

    Parameters
    ----------
    a : 1D or 2D np.ndarray, shape (n, p) or (p,)
        Parameters for the Dirichlet distribution of each sample.
    size : int, optional
        Number of samples to generate. Default is 1.

    Returns
    -------
    ndarray, shape (n, p)
        An array of Dirichlet samples, each normalized to sum to 1.
    """
    n = size
    if a.ndim == 1:
        # a is a 1D array
        p = a.shape[0]
    elif a.ndim == 2:
        # a is a 2D array
        p = a.shape[1]
        if n != a.shape[0]:
            raise ValueError(
                "if a is a 2D array, \
            n should be equal to the number of rows of a"
            )
    else:
        raise ValueError("a must be either a 1D or 2D array")
    x = np.random.gamma(a, size=(n, p))
    sm = np.sum(x, axis=1)
    return x / sm[:, None]


def pdf_dirichlet(x, a):
    """
    Evaluate the probability density function of a `Dirichlet distribution <https://en.wikipedia.org/wiki/Dirichlet_distribution>`_ with parameter `a`, at each point represented as a row in the data matrix `x`.

    Parameters
    ----------
    x : ndarray, shape (n, d) or (d,)
        Data matrix where each row is a point on the (d-1)-dimensional simplex.
    a : ndarray, shape (d,)
        Parameters for the Dirichlet distribution.

    Returns
    -------
    ndarray, shape (n,)
        Probability density at each row of `x` for the given Dirichlet
        distribution.
    """
    if x.ndim == 1:
        x = x[np.newaxis, :]

    assert np.allclose(np.sum(x, axis=1), 1), "Each row of x must sum to 1"
    assert x.shape[1] == len(a), (
        "Each row of x and a must have \
                                      the same length"
    )

    log_numerator = np.sum((a - 1) * np.log(x), axis=1)
    log_denominator = np.sum(loggamma(a)) - loggamma(np.sum(a))

    log_pdf = log_numerator - log_denominator
    return np.exp(log_pdf)


def pdf_dirimix(x, Mu, wei, lnu):
    """
    Density function of a Dirichlet mixture distribution on the (d-1)-dimensional simplex evaluated at each
    row in the data matrix `x`.

    Each Dirichlet mixture component has weight given by the jth entry
    of `wei`, and Dirichlet parameter (see :func:`pdf_dirichlet`) :math:`a_j=\\nu_j \\mu_j`, where
    :math:`\\nu_j = \\exp(\\textrm{lnu}_j)`, :math:`\\mu_j` is the jth
    row of matrix `Mu`, :math:`\\textrm{lnu}_j` is the jth entry of `lnu`.
    The Dirichlet mixture density writes, for `x` in the unit simplex,

    .. math::

        f(x, \\textrm{Mu}, \\textrm{wei}, \\textrm{lnu}) = \\sum_{j=0}^{k-1} \
    \\textrm{wei}[j] \\textrm{dirichlet}(x | \\textrm{Mu}[j, :], \\exp(\\textrm{lnu}[j])),

    where "dirichlet" is the Dirichlet density with parameter \
    :math:`a=\\nu\\mu`,
    
    .. math::

        \\textrm{dirichlet}(x | \\mu, \\nu) = \\frac{\\Gamma(\\nu)}{\\prod_{i=1}^d \\Gamma(\\nu \\mu_i)}\\prod_{i=1}^d x_i^{\\nu \\mu_i - 1}.

    

    Parameters
    ----------
    x : ndarray, shape (n, d) or (d,)
        Data matrix where each row is a point on the (d-1)-dimensional simplex.
    Mu : ndarray, shape (k, d)
        Matrix of means for the Dirichlet concentration parameters.
        Each row must contain non-negative entries and sum to one.
        A moments constraint arises from standardization to unit Pareto margins.
        This constraint is that

        .. math::

            \\sum_{j=1}^{k} w_j \\mu_j = \\left(\\frac{1}{d}, \\ldots, \\frac{1}{d}\\right).

        No warning nor error is issued if that constraint is not satisfied, to encompass more general situations.
    wei : ndarray, shape (k,)
        Vector of weights for each mixture component. Nonnegative entries, summing to one.
    lnu : ndarray, shape (k,)
        Vector of log-scales for each mixture component.

    
    
    Returns
    -------
    ndarray, shape (n,)
        Density at each row of `x` on the (d-1)-dimensional simplex.
    """
    if x.ndim == 1:
        x = x[np.newaxis, :]

    k = len(wei)
    density = np.zeros(x.shape[0])
    for i in range(k):
        a = Mu[i, :] * np.exp(lnu[i])
        density += wei[i] * pdf_dirichlet(x, a)

    return density


def gen_dirimix(Mu, wei, lnu, size=1):
    """
    Generate angular samples from a mixture of Dirichlet distributions \
    described in :func:`pdf_dirimix`

    Parameters
    -----------
    size : int (optional)
        Number of samples to generate.
    Mu : ndarray, shape (k, p)
        Matrix of means for the Dirichlet components.
    wei : ndarray, shape (k,)
        Vector of weights for each mixture component.
    lnu : ndarray, shape (k,)
        Vector of log-scales for each mixture component.

    Returns
    --------
    ndarray, shape (size, p)
        Array of Dirichlet samples generated from the mixture.
    """
    n = size
    k, p = Mu.shape
    if len(lnu) != k or len(wei) != k:
        raise ValueError("Length of lnu and wei must be equal to k")

    u = np.random.rand(n)
    cum_wei = np.cumsum(wei)

    ms = np.array(
        [
            np.where(cum_wei <= u[i])[0].max() + 1 if np.any(cum_wei <= u[i]) else 0
            for i in range(n)
        ]
    )

    matpars = np.array([Mu[ms[i], :] * np.exp(lnu[ms[i]]) for i in range(n)])
    return gen_dirichlet(a=matpars, size=n)


def plot_pdf_dirimix_2D(Mu, wei, lnu, n_points=500):
    """
    Plot the Dirichlet mixture density on the 1-D simplex in \
ambient dimension 2, see :func:`pdf_dirimix`.

    Parameters
    -----------
    Mu : ndarray, shape (2, k)
        Matrix of means for the Dirichlet concentration parameters.
    lnu : ndarray, shape (k,)
        Vector of log-scales for each mixture component.
    wei : ndarray, shape (k,)
        Vector of weights for each mixture component.
    n_points : int, optional
        Number of points to use for evaluating the density. Default is 500.
    """
    x_values1 = np.linspace(10 ** (-5), 1 - 10 ** (-5), n_points)
    x = np.column_stack((x_values1, 1 - x_values1))
    density_values = pdf_dirimix(x, Mu, wei, lnu)
    plt.figure(figsize=(8, 6))
    plt.plot(x_values1, density_values, label="Mixture Density", color="blue")
    plt.fill_between(x_values1, density_values, alpha=0.3, color="blue")
    plt.title("Mixture Density on the 1-D Simplex")
    plt.xlabel("x[1] first component")
    plt.ylabel("Density")
    plt.grid(True)
    plt.legend()
    plt.show()


def plot_pdf_dirimix_3D(Mu, wei, lnu, n_points=500):
    """
    Plot the mixture density on the 2D simplex in ambient dimension 3, \
see :func:`pdf_dirimix`.

    Parameters
    -----------
    Mu : ndarray, shape (3, k)
        Matrix of means for the Dirichlet concentration parameters.
    lnu : ndarray, shape (k,)
        Vector of log-scales for each mixture component.
    wei : ndarray, shape (k,)
        Vector of weights for each mixture component.
    n_points : int, optional
        Number of points to use for evaluating the density. Default is 500.
    """
    x1_values = np.linspace(10 ** (-5), 1 - 10 ** (-5), n_points)
    x2_values = np.linspace(10 ** (-5), 1 - 10 ** (-5), n_points)

    X1, X2 = np.meshgrid(x1_values, x2_values)
    X1 = X1.flatten()
    X2 = X2.flatten()

    valid_points = X1 + X2 <= 1 - 10 ** (-5)
    X1 = X1[valid_points]
    X2 = X2[valid_points]
    X_full = np.column_stack((X1, X2, 1 - X1 - X2))
    density_values = pdf_dirimix(X_full, Mu, wei, lnu)

    plt.figure(figsize=(8, 6))
    plt.tricontourf(X1, X2, density_values, 20, cmap="viridis")
    plt.colorbar(label="Density")
    plt.title("Mixture Density on the 2D Simplex")
    plt.xlabel("x1 (first component)")
    plt.ylabel("x2 (second component)")

    plt.plot([0, 1, 0], [1, 0, 0], "k--", lw=2)
    plt.grid(True)
    plt.show()


# ##############################################
# ## Multivariate symmetric / asymmetric logistic Model
# ##############################################
def gen_PositiveStable(alpha, size=1):
    """
    Generate positive stable random variables, useful for generating \
    Multivariate Logistic variables, see :func:`gen_multilog`.

    Parameters
    -----------
    size : int (optional)
        Sample size.
    alpha : float
        Dependence parameter.
  
    Returns
    --------
    ndarray
        Sample of positive stable random variables.
    """
    U = np.pi * np.random.random(size)
    #    U = stat.uniform.rvs(0, np.pi, size=size, random_state=seed)
    W = np.random.exponential(scale=1, size=size)
    # stat.expon.rvs(size=size, random_state=seed)
    Term_1 = (np.sin((1 - alpha) * U) / W) ** ((1 - alpha) / alpha)
    Term_2 = np.sin(alpha * U) / (np.sin(U) ** (1 / alpha))
    return Term_1 * Term_2


def gen_multilog(dim, alpha, size=1):
    """
    Generate multivariate symmetric logistic random variables via Algorithm 2.1 in [1].

    [1]: Stephenson, A. (2003). Simulating multivariate extreme value distributions of logistic type.\
    Extremes, 6, 49-59.

    Parameters
    -----------
    size : int (optional)
        Sample size.
    dim : int
        Dimension.
    alpha : float
        Dependence parameter.
  
    Returns
    --------
    ndarray
        Sample of multivariate logistic random variables.
    """
    W = np.random.exponential(scale=1, size=(size, dim))
    #    W = stat.expon.rvs(size=(size, dim), random_state=seed)
    S = gen_PositiveStable(alpha=alpha, size=size)
    Result = np.zeros((size, dim))
    for ii in range(dim):
        Result[:, ii] = (S / W[:, ii]) ** alpha
    return Result


# ###########################################
# ### Prediction of a missing component
# ###########################################


def transform_target_lin(y, X, norm_func):
    """
    Transform the target vector to achieve approximate independence from the
    norm of `X` and stabilize the learning algorithms by preventing large target values

    This function rescales the original target vector `y` into
    `y' = y / ||X||` to mitigate the influence of the magnitude of `X` on `y`,
    particularly when `||X||` is large. This transformation is useful for
    predicting missing components in heavy-tailed multivariate data.

    Parameters
    ----------
    y : array-like
        The original target vector to be transformed.
    X : array-like
        The matrix whose norm is used to rescale `y`.
    norm_func : callable
        A function that computes the norm of `X`.

    Returns
    -------
    y1 : array-like
        The rescaled target vector.
    """
    norm_X = norm_func(X)
    y1 = y / norm_X
    return y1


def inv_transform_target_lin(y, X, norm_func):
    """
    Inverse transform the rescaled target vector to its original scale.

    This function performs the inverse operation of :func:`transform_target_lin`,
    converting the rescaled target vector back to its original scale by
    multiplying it by the norm of `X`.

    Parameters
    ----------
    y : array-like
        The rescaled target vector to be inverse-transformed.
    X : array-like
        The matrix whose norm was used to rescale `y`.
    norm_func : callable
        The same norm function used in `transform_target_lin`.

    Returns
    -------
    y_orig : array-like
        The original target vector.
    """
    norm_X = norm_func(X)
    y_orig = y * norm_X
    return y_orig


def transform_target_nonlin(y, X, norm_order):
    """Non-linearly transform the target vector to achieve \
    approximate independence from the norm of `X` and stabilize the
    learning algorithms by preventing large target values

    This function rescales the original target vector `y` into `y' = y\
    / ||[X, y]||_q` to mitigate the influence of the magnitude of `X`\
    on `y`, particularly when `||[X, y]||_q` is large. This\
    transformation is useful for predicting missing components in
    heavy-tailed multivariate data.

    Parameters
    ----------
    y : array-like
        The original target vector to be transformed.
    X : array-like
        The matrix whose norm, combined with `y`, is used to rescale `y`.
    norm_order : int
        The order of the norm to be used for the transformation.

    Returns
    -------
    array-like
        The rescaled target vector.

    """
    q = norm_order
    joint = np.hstack((X, y.reshape(-1, 1)))
    norm_full = np.linalg.norm(joint, ord=q, axis=1)
    return y / norm_full


def inv_transform_target_nonlin(y, X, norm_order):
    """
    Inverse transform the rescaled target vector to its original scale in a nonlinear fashion.

    This function performs the inverse operation of
    :func:`transform_target_nonlin`,
    converting the rescaled target vector back to its original scale.

    Parameters
    ----------
    y : array-like
        The rescaled target vector to be inverse-transformed.
    X : array-like
        The matrix whose norm was used to rescale `y`.
    norm_order : int
        The order of the norm used in the transformation.

    Returns
    -------
    array-like
        The original target vector.
    """
    q = norm_order
    norm_x = np.linalg.norm(X, ord=q, axis=1)
    predicted_x = y / (1 - y**q) ** (1 / q) * norm_x
    return predicted_x


def gen_rv_dirimix(
    alpha=1,
    Mu=np.array([[0.5, 0.5]]),
    wei=np.array([1]),
    lnu=np.array([2]),
    scale_weight_noise=1,
    index_weight_noise=None,
    Mu_bulk=None,
    wei_bulk=None,
    lnu_bulk=None,
    size=1,
):
    """Generate `n` points according to a multivariate regularly varying \
    distribution (heavy-tailed) with regular variation `alpha`.

    Each point is generated as `X_i = R_i * Theta_i`, where R_i and
    Theta_i are not independent in the bulk.

    The noise influence decreases proportionally to
    `(scale_weight_noise/R)**(index_weight_noise)`, where `R` is the 1-norm of the sample point.
    This means that as the radius `R` increases, the noise's effect diminishes, controlled by the
    `scale_weight_noise` and `index_weight_noise` parameters.

    Theta is generated as a weighted average of two components:
    - A "bulk" component, which dominates for small to moderate radii.
    - A "tail" component, which becomes more significant as the radius increases.

    Parameters
    -----------
    size : int, optional
        Number of samples to generate. Default is 1.
    alpha : float, optional
        Shape parameter of the Pareto distribution. Default is 1.
    Mu : ndarray, shape (k,d)
        Means for the Dirichlet mixture components.
    wei : ndarray, shape (k,)
        Weights for each mixture component.
    lnu : ndarray, shape (k,)
        Log-scale parameters for the Dirichlet mixture components.
    Mu_bulk : ndarray, shape (k,d), optional
        Parameter matrix of same dimension as Mu, ruling the angular distribution for small to moderate radii.
    scale_weight_noise : float, optional
        Scaling factor (>0) for the angular noise. It controls the overall magnitude of the noise added to the angular component.
        A larger value increases the noise influence, making points more dispersed, while a smaller value reduces noise effect.
        Default is 1.
    index_weight_noise : float, optional
        Exponent ruling how the noise vanishes for large R. It determines how quickly the noise's influence diminishes as the radius `R` increases.
        A higher value makes the noise vanish more rapidly, resulting in less noisy points for larger radii.
        Default is -alpha.
    wei_bulk : ndarray, shape (k,), optional
        Weights for each mixture component in the bulk distribution.
    lnu_bulk : ndarray, shape (k,), optional
        Log-scale parameters for the Dirichlet mixture components in the bulk distribution.

    Returns
    --------
    ndarray, shape (n, d)
        Array of points generated from the Pareto and Dirichlet mixture.

    """
    n = size
    if index_weight_noise is None:
        index_weight_noise = alpha

    R = np.random.pareto(alpha, size=n) + 1
    Theta = gen_dirimix(Mu=Mu, wei=wei, lnu=lnu, size=n)
    dim = np.shape(Mu)[1]
    # k_comp = np.shape(Mu)[0]
    # if Mu_bulk is None:
    #     Mu_bulk = (1 - Mu)/(dim - 1)
    if Mu_bulk is None:
        Mu_bulk_0 = np.maximum((2 / dim - Mu), 10 ** (-5))
        r_0 = np.sum(Mu_bulk_0, axis=1)
        Mu_bulk = Mu_bulk_0 / r_0.reshape(-1, 1)

    if wei_bulk is None:
        wei_bulk = wei  # np.ones(k_comp)/k_comp

    if lnu_bulk is None:
        lnu_bulk = lnu[::-1]

    Noise = gen_dirimix(Mu=Mu_bulk, wei=wei_bulk, lnu=lnu_bulk, size=n)
    # gen_dirichlet(n, ) np.ones((n, np.shape(Mu)[0])))
    w = np.minimum(1, (scale_weight_noise / R) ** (index_weight_noise))
    newTheta = (1 - w[:, np.newaxis]) * Theta + w[:, np.newaxis] * Noise
    X = R[:, np.newaxis] * newTheta

    return X


def gen_classif_data_diriClasses(
    mu0=np.array([0.7, 0.3]), lnu=None, alpha=4, index_weight_noise=1, size=10
):
    """
    Generate synthetic classification data with two classes using a Dirichlet mixture model.

    This function generates synthetic data for binary classification tasks. It creates two classes
    of data points using a Dirichlet mixture model, where each class is characterized by different
    mean parameters. The data points are generated using the `gen_rv_dirimix` function, which
    combines Pareto-distributed radii with Dirichlet-distributed angular components.

    Parameters
    -----------
    mu0 : ndarray, shape (d,), optional
        Mean parameter for the first class. Default is np.array([0.7, 0.3]).
    lnu : ndarray, shape (k,), optional
        Log-scale parameters for the Dirichlet mixture components. If not provided, it is calculated
        based on the mean parameters.
    alpha : float, optional
        Shape parameter of the Pareto distribution. Default is 4.
    index_weight_noise : float, optional
        Exponent ruling how the noise vanishes for large R. Default is 1.
    size : int, optional
        Total number of samples to generate. Default is 10.

    Returns
    --------
    X : ndarray, shape (n, d)
        Array of generated data points.
    y : ndarray, shape (n,)
        Array of class labels corresponding to the data points.

    Notes
    -----
    The function first normalizes the mean parameters and weights for the Dirichlet mixture model.
    It then generates data points for each class using the `gen_rv_dirimix` function, combining them
    into a single dataset. The class labels are randomly permuted to ensure a balanced dataset.
    """
    mu1 = 1 - mu0
    Mu_temp = np.array([mu0, mu1])
    wei_temp = 0.5 * np.ones(2)
    Mu, wei = normalize_param_dirimix(Mu_temp, wei_temp)
    if lnu is None:
        lnu = np.log(2 / np.min(Mu, axis=1))
    size0 = int(size * wei[0])
    size1 = int(size * wei[1])
    Mu0 = Mu[0, :].reshape(1, -1)
    Mu1 = Mu[1, :].reshape(1, -1)
    # Dim = np.shape(Mu)[1]
    data0 = gen_rv_dirimix(
        alpha,
        Mu0,
        wei=np.array([1]),
        lnu=np.array([lnu[0]]),
        scale_weight_noise=10 ** (1 / alpha),  # np.sqrt(Dim),
        index_weight_noise=index_weight_noise,
        size=size0,
    )
    data1 = gen_rv_dirimix(
        alpha,
        Mu1,
        wei=np.array([1]),
        lnu=np.array([lnu[1]]),
        scale_weight_noise=10 ** (1 / alpha),  # np.sqrt(Dim),
        index_weight_noise=index_weight_noise,
        size=size1,
    )
    y = np.vstack(
        (np.zeros(size0).reshape(-1, 1), np.ones(size1).reshape(-1, 1))
    ).flatten()
    X = np.vstack((data0, data1))
    permut = np.random.permutation(size0 + size1)
    X = X[permut, :]
    y = y[permut]
    return X, y


# ## target generation for regression models:
# specific instance of additive noise nodel considered in Huet et al.
def tail_reg_fun_default(angle):
    """Default tail regression function for generating target values.

    This function is used internally as an argument for
    :func:`gen_target_CovariateRV`.  The regression function in the
    tail is defined in a linear model, y = < beta, angle >. `angle` is
    meant to be the angular component of the original input, and
    `beta` is a hard-set vectors with the first half of its entries
    equal to 10, and  the rest equal to 0.1.



    Parameters
    ----------
    angle : np.ndarray
        A 1D or 2D array representing angular components of the data.
        Each element corresponds to an angle in the covariate space. If
        `angle` is a 2D array, each row represents a different observation.

    Returns
    -------
    np.ndarray
        The result of the dot product between the `angle` array and the
        vector `beta`. This output represents the tail regression values
        for the input angles, which can be used to generate target values
        in a tail regression model.
    """
    if angle.ndim == 1:
        # angle is a 1D array
        p = angle.shape[0]
    elif angle.ndim == 2:
        # angle is a 2D array
        p = angle.shape[1]
    p1 = int(p / 2)
    beta = np.concatenate([10 * np.ones(p1), 0.1 * np.ones(p - p1)])
    return np.dot(angle, beta)


def bulk_reg_fun_default(angle):
    """Default bulk regression function for generating target values.

    This function is used internally as an argument for
    :func:`gen_target_CovariateRV`.  The regression function in the
    bulk is defined in a linear model, y = < beta, angle >. `angle` is
    meant to be the angular component of the original input, and
    `beta` is a hard-set vectors with the first half of its entries
    equal to 0.1, and  the rest equal to 10.

    Parameters
    ----------
    angle : np.ndarray
        A 1D or 2D array representing angular components of the data.
        Each element corresponds to an angle in the covariate space. If
        `angle` is a 2D array, each row represents a different
        observation.

    Returns
    -------
    np.ndarray
        The result of the dot product between the `angle` array and the
        vector `beta`. This output represents the bulk regression values
        for the input angles, which can be used to generate target values
        in a regression model.

    """
    if angle.ndim == 1:
        # angle is a 1D array
        p = angle.shape[0]
    elif angle.ndim == 2:
        # angle is a 2D array
        p = angle.shape[1]
    p1 = int(p / 2)
    beta = np.concatenate([0.1 * np.ones(p1), 10 * np.ones(p - p1)])
    return np.dot(angle, beta)


def bulk_decay_fun_default(radius, rv_index):
    """Default bulk decay function for generating continuous target values
    in a regression model. 

    This function applies a decay transformation to the input radii,
    and returns : `1 / (radius) ** (rv_index)`

    The ouput is used in function :func:`gen_target_CovariateRV` as a\
     weight to be attributed to the bulk regression function, and the\
     final regression function in :func:`gen_target_CovariateRV` is a\
     weighted average between a bulk and a tail component.
    
    Parameters
    ----------
    radius : np.ndarray
        A 1D array representing the radial distances. Each element
        corresponds to the radius of a data point in the covariate space.

    rv_index : float
        A positive float that determines the rate of decay. Higher values
        result in faster decay. This parameter allows for adjustment of
        the decay function's sensitivity to changes in radius.

    Returns
    -------
    np.ndarray
        An array containing the result of applying the decay function to
        each element in the `radius` array. The output represents the
        decayed values, which can be used to weight the contributions of
        different data points in a regression model.

    """
    return 1 / (radius) ** (rv_index)


def gen_target_CovariateRV(
    X,
    tail_reg_fun=tail_reg_fun_default,
    bulk_reg_fun=bulk_reg_fun_default,
    bulk_decay_fun=bulk_decay_fun_default,
    param_decay_fun=2,
):
    """Generate target values for covariate random variables.

    This function generates target values for a given set of covariate data
    by combining tail and bulk regression functions with a decay function.
    The resulting target values are influenced by both the tail and bulk
    regression functions, weighted by the decay function.

    Parameters
    ----------
    X : np.ndarray
        A 2D array representing covariate data. Each row corresponds to a
        different observation, and each column corresponds to a different
        covariate.

    tail_reg_fun : function, optional
        A function that computes the tail regression values based on the
        angular components of the covariate data. Default is
        `tail_reg_fun_default`.

    bulk_reg_fun : function, optional
        A function that computes the bulk regression values based on the
        angular components of the covariate data. Default is
        `bulk_reg_fun_default`.

    bulk_decay_fun : function, optional.
        A function that computes the
        weight of the bulk component of the regression function,
        based on the radial components of the covariate data. Default
        is `bulk_decay_fun_default`.

    param_decay_fun : any, optional.
        An additional parameter to be
        passed to `bulk_decay_fun`. If `bulk_decay_fun_default` is
        used, then `param_decay_fun` is the regular variation index of
        the weight, as a function of the radius. Default is 2. Larger
        values induce a more rapidly decaying bulk weight.

    Returns
    -------
    y : np.ndarray
        An array of generated target values corresponding to the input
        covariate data. The target values are computed as a weighted
        combination of the tail and bulk regression values, with added
        Gaussian noise.

    """
    rad = np.linalg.norm(X, axis=1)
    ang = X / rad[:, np.newaxis]
    n = len(rad)
    noise = np.random.normal(loc=0, scale=0.01, size=n)
    tail_mean = tail_reg_fun(ang).flatten()
    bulk_mean = bulk_reg_fun(ang).flatten()
    bulk_weight = bulk_decay_fun(rad, param_decay_fun).flatten()
    ystar = (1 - bulk_weight) * tail_mean + bulk_weight * bulk_mean
    y = ystar + noise
    return y


# ## toy example for functional PCA
def gen_rv_functional_data(
    num_samples, grid, alpha, alphanoise, scalenoise=5, om1=1, om2=2, om3=3, om4=4
):
    """
    Generate random regularly varying functions on a grid using
    Pareto-distributed coefficients and noise components.

    Parameters
    -----------
    num_samples : int
        Number of samples to generate.
    grid : array-like
        1D array representing the grid over which the basis functions are
        evaluated.
    alpha : float
        Shape parameter for the 'signal' Pareto-distributed variables.
    alphanoise : float
        Shape parameter for the noise Pareto-distributed variables.
    scalenoise : float, optional
        Scaling factor for the noise components. Default is 5.
    om1, om2, om3, om4 : float, optional
        Frequencies used in the sine and cosine basis functions.
        Defaults are 1, 2, 3, 4.

    Returns
    --------
    ndarray
        A 2D NumPy array of shape `(num_samples, len(grid))`, where each row
    represents a generated functional data sample.
    """
    a1 = (np.random.pareto(alpha, size=num_samples) + 1) * (
        2 * (np.random.uniform(size=num_samples) > 0.5) - 1
    )
    a2 = (np.random.pareto(alpha, size=num_samples) + 1) * (
        2 * (np.random.uniform(size=num_samples) > 0.5) - 1
    )
    a3 = (
        scalenoise
        * (np.random.pareto(alphanoise, size=num_samples) + 1)
        * (2 * (np.random.uniform(size=num_samples) > 0.5) - 1)
    )
    a4 = (
        scalenoise
        * (np.random.pareto(alphanoise, size=num_samples) + 1)
        * (2 * (np.random.uniform(size=num_samples) > 0.5) - 1)
    )

    result = (
        np.dot(a1[:, None], (1 + np.sin(2 * np.pi * om1 * grid))[None, :])
        + np.dot(a2[:, None], (1 + np.cos(2 * np.pi * om2 * grid))[None, :])
        + np.dot(a3[:, None], (1 + np.sin(2 * np.pi * om3 * grid))[None, :])
        + np.dot(a4[:, None], (1 + np.cos(2 * np.pi * om4 * grid))[None, :])
    )

    return result


def gen_rv_functional_data_gaussianNoise(
    num_samples, grid, alpha, sd, scalenoise=5, om1=1, om2=2, om3=3, om4=4, om5=5, om6=6
):
    """
    Generate random regularly varying functions on a grid using
    Pareto-distributed coefficients and Gaussian noise components.

    Parameters
    -----------
    num_samples : int
        Number of samples to generate.
    grid : array-like
        1D array representing the grid over which the basis functions are
        evaluated.
    alpha : float
        Shape parameter for the Pareto-distributed variables.
    sd : float
        Standard deviation for the Gaussian noise components.
    scalenoise : float, optional
        Scaling factor for the noise components. Default is 5.
    om1, om2, om3, om4, om5, om6 : float, optional
        Frequencies used in the sine and cosine basis functions.
        Defaults are 1, 2, 3, 4, 5, 6.

    Returns
    --------
    ndarray
        A 2D NumPy array of shape `(num_samples, len(grid))`, where each row
        represents a generated data sample.
    """
    a1 = np.random.pareto(alpha, size=num_samples) + 1
    a2 = 0.8 * (np.random.pareto(alpha, size=num_samples) + 1)
    a3 = scalenoise * np.random.normal(0, sd, size=num_samples)
    a4 = 0.8 * scalenoise * np.random.normal(0, sd, size=num_samples)
    a5 = 0.6 * scalenoise * np.random.normal(0, sd, size=num_samples)
    a6 = 0.4 * scalenoise * np.random.normal(0, sd, size=num_samples)

    result = (
        np.dot(a1[:, None], (np.sqrt(2) * np.sin(2 * np.pi * om1 * grid))[None, :])
        + np.dot(a2[:, None], (np.sqrt(2) * np.cos(2 * np.pi * om2 * grid))[None, :])
        + np.dot(a3[:, None], (np.sqrt(2) * np.sin(2 * np.pi * om3 * grid))[None, :])
        + np.dot(a4[:, None], (np.sqrt(2) * np.cos(2 * np.pi * om4 * grid))[None, :])
        + np.dot(a5[:, None], (np.sqrt(2) * np.sin(2 * np.pi * om5 * grid))[None, :])
        + np.dot(a6[:, None], (np.sqrt(2) * np.cos(2 * np.pi * om6 * grid))[None, :])
    )

    return result


# ######
# subface generation for feature clustering
# #####
# #################
# Generate subfaces #
# #################


def gen_subfaces(
    dimension,
    num_subfaces,
    max_size=8,
    p_geometric=0.25,
    prevent_inclusions=True,
    # include_singletons=True,
    seed=None,
):
    """
    Generates a list of random subsets (subfaces) of the set {1, ..., dimension}.

    This function is used for data generation in feature clustering algorithms.
    It creates random subsets of a specified dimension, ensuring that each subset
    meets certain size and inclusion criteria.

    Parameters
    -----------
    dimension : int
        Dimensionality of the ambient space, i.e., the range of indices for the subsets.
    num_subfaces : int
        Number of subfaces of size >= 2 to generate.
    max_size : int, optional
        Maximum size of a subface. Default is 8.
    p_geometric : float, optional
        Parameter for the geometric distribution ruling the size of subfaces.
        Default is 0.25.
    prevent_inclusions : bool, optional
        If True, ensures that no subface is a subset or superset of another.
        Default is True.
    seed : int, optional
        Seed for random number generation to ensure reproducibility.
        Default is None.

    Returns
    --------
    list
        List of generated subfaces, where each subface is represented as a list of integers.

    Notes
    -----
    The function uses a geometric distribution to determine the size of each subface.
    It ensures that each feature (index) is included in at least one subface to allow
    for standardized components in the associated random vector (dataset).
    """
    if seed is not None:
        random.seed(seed)
        np.random.seed(seed)

    # create a matrix of subfaces
    subfaces = np.zeros((num_subfaces, dimension))
    # subface_size = min(np.random.geometric(p_geometric) + 1, max_size)
    # subfaces[0, random.sample(range(dimension), subface_size)] = 1
    count = 0
    loop_count = 0
    while count < num_subfaces and loop_count < 1e4:
        subface_size = min(np.random.geometric(p_geometric) + 1, max_size)
        subface = np.zeros(dimension)
        subface_indices = random.sample(range(dimension), subface_size)
        subface[subface_indices] = 1
        if count >= 1:
            valid_cond = (not prevent_inclusions) or (
                # subface is not a subset of one of subfaces
                (np.sum(np.prod(subfaces[:count] * subface == subface, axis=1)) == 0)
                and (
                    # subface is not a superset either
                    np.sum(
                        np.prod(subfaces[:count] * subface == subfaces[:count], axis=1)
                    )
                    == 0
                )
            )
        else:
            valid_cond = True
        if valid_cond:
            subfaces[count, subface_indices] = 1
            count += 1
        loop_count += 1

    idkeep = np.sum(subfaces, axis=1) > 0
    subfaces = subfaces[idkeep]
    # convert the subfaces matrix to a list
    subfaces_list = [list(np.nonzero(f)[0]) for f in subfaces]
    # features = list({int(j) for subface in subfaces_list for j in subface})

    # Each feature must be in at least one subface because otherwise
    # the associated random vector (dataset) cannot have standardized
    # components (columns). The last step is to add potentially
    # missing features:

    missing_features = list(
        set(range(dimension)) - {j for subface in subfaces_list for j in subface}
    )
    # singletons = []

    if missing_features:
        #       if include_singletons:
        # singletons = [[int(j)] for j in missing_features]
        for j in missing_features:
            subfaces_list.append([int(j)])
        # else:
        #     if len(missing_features) > 1:
        #         subfaces_list.append(missing_features)
        #     if len(missing_features) == 1:
        #         missing_features.append(list(set(range(dimension)) -
        #                                      set(missing_features))[0])
        #         subfaces_list.append(missing_features)

    converted_subfaces = [[int(item) for item in sublist] for sublist in subfaces_list]

    # if include_singletons:
    #     return converted_subfaces, features, singletons
    return converted_subfaces



# Author: Anne Sabourin
# Date: 2025
"""
Tools for Extreme Value Analysis in the MLExtreme Package

This submodule provides essential tools for conducting Extreme Value Analysis (EVA) in various contexts and examples within the MLExtreme package. It includes functionalities for rank transformation and a specialized tool for selecting extreme thresholds based on distance covariance tests. These tools are designed to support both multivariate supervised and unsupervised settings, facilitating robust analysis and decision-making in extreme value scenarios.
"""

import numpy as np
import matplotlib.pyplot as plt
import warnings


def round_signif(numbers, digits):
    """
    Round each number in a list, NumPy array, or a single float to a specified
    number of significant digits.

    Parameters
    ----------
    numbers : float, list, or np.ndarray
        The number, list, or array of numbers to round.

    digits : int
        The number of significant digits to round to.

    Returns
    -------
    float, list, or np.ndarray
        The rounded numbers in the same format as the input.

    Examples
    --------
    >>> import numpy as np
    >>> statistics_array = np.array([0.12345, 6.789, 0.00567])
    >>> statistics_list = [0.12345, 6.789, 0.00567]
    >>> single_float = 0.12345

    Round each element to two significant digits

    >>> rounded_statistics_array = round_signif(statistics_array, 2)
    >>> rounded_statistics_list = round_signif(statistics_list, 2)
    >>> rounded_single_float = round_signif(single_float, 2)

    Print the rounded statistics

    >>> print(f'Statistics (array): {rounded_statistics_array}')
    Statistics (array): [0.12 6.8 0.0057]
    >>> print(f'Statistics (list): {rounded_statistics_list}')
    Statistics (list): [0.12, 6.8, 0.0057]
    >>> print(f'Single float: {rounded_single_float}')
    Single float: 0.12
    """
    if isinstance(numbers, float):
        # Round a single float
        return (
            float(round(numbers, digits - int(np.floor(np.log10(abs(numbers)))) - 1))
            if numbers != 0
            else 0.0
        )
    if isinstance(numbers, np.ndarray):
        # Apply rounding element-wise and return as a NumPy array
        return np.array(
            [
                float(round(num, digits - int(np.floor(np.log10(abs(num)))) - 1))
                if num != 0
                else 0.0
                for num in numbers
            ]
        )
    if isinstance(numbers, list):
        # Apply rounding element-wise and return as a list
        return [
            float(round(num, digits - int(np.floor(np.log10(abs(num)))) - 1))
            if num != 0
            else 0.0
            for num in numbers
        ]
    else:
        raise TypeError("Input must be a float, list, or NumPy array of floats.")


def hill_estimator(x, k):
    """
    Hill estimator for the tail index based on the data `x`, using the
    `k` largest order statistics.

    Parameters
    ----------
    x : array-like
        The input data sample from which to estimate the tail index.
    k : int, >= 2
        The number of extreme values to consider for the estimation.

    Returns
    -------
    dict
        A dictionary containing the Hill estimator value ('estimate') and its
        standard deviation ('sdev').
    """
    sorted_data = np.sort(x)[::-1]  # Sort data in decreasing order
    hill_estimate = 1 / (k - 1) * np.sum(np.log(sorted_data[:k] / sorted_data[k - 1]))
    standard_deviation = hill_estimate / np.sqrt(k - 1)
    return {"estimate": hill_estimate, "sdev": standard_deviation}


def rank_transform(x_raw):
    """
    Standardize each column of the input matrix using a Pareto-based
    rank transformation.

    This function transforms each element in the input matrix by assigning
    a rank based on its position within its column and then applying a
    Pareto transformation. The transformation is defined as:

    .. math::

        v_{\\text{rank},ij} = \\frac{n_{\\text{sample}} + 1}{\\text{rank}(x_{\\text{raw},ij}) + 1} = \\frac{1}{1 - \\frac{n}{n+1} F_{\\text{emp}}(x_{\\text{raw},ij})}

    where :math:`\\text{rank}(x_{\\text{raw},ij})` is the rank of the element
    :math:`x_{\\text{raw},ij}` within its column in decreasing order and
    :math:`F_{\\text{emp}}` is the usual empirical CDF.

    Parameters
    ----------
    x_raw : numpy.ndarray
        A 2D array where each column represents a feature and each row
        represents a sample. The function assumes that the input is a
        numerical matrix.

    Returns
    -------
    numpy.ndarray
        A transformed matrix of the same shape as `x_raw`, where each element
        has been standardized using the Pareto-based rank transformation.

    Example
    -------
    >>> import numpy as np
    >>> x_raw = np.array([[10, 20], [30, 40], [50, 60]])
    >>> rank_transform(x_raw)
    array([[1.33333333, 1.33333333],
           [2.        , 2.        ],
           [4.        , 4.        ]])
    """
    n_sample, n_dim = x_raw.shape
    mat_rank = np.argsort(x_raw, axis=0)[::-1]
    x_rank = np.zeros((n_sample, n_dim))
    for j in range(n_dim):
        x_rank[mat_rank[:, j], j] = np.arange(n_sample) + 1
    v_rank = (n_sample + 1) / x_rank

    return v_rank


def rank_transform_test(x_train, x_test):
    """
    Transform each column in `x_test` to approximate a unit Pareto distribution
    based on the empirical cumulative distribution function (ECDF) resulting
    from the corresponding column in `x_train`.

    Parameters
    ----------
    x_train : numpy.ndarray
        A 2D array of shape (n, d) representing the training data.
    x_test : numpy.ndarray
        A 2D array of shape (m, d) representing the test data to be transformed.

    Returns
    -------
    numpy.ndarray
        A transformed version of `x_test` of shape (m, d), where each element is
        transformed using the formula:

        .. math::

            \\frac{1}{1 - \\frac{n}{n+1} F_{\\text{emp}}(x_{\\text{test},ij})}

    Example
    -------
    >>> import numpy as np
    >>> x_train = np.array([[1, 2], [3, 4], [5, 6]])
    >>> x_test = np.array([[2, 3], [4, 5]])
    >>> rank_transform_test(x_train, x_test)
    array([[1.33333333, 1.33333333],
           [2.        , 2.        ]])
    """
    n_samples, n_dim = x_train.shape
    m_samples, _ = x_test.shape

    # Initialize the transformed array
    x_transf = np.zeros((m_samples, n_dim))

    # Iterate over each column
    for j in range(n_dim):
        # Compute   ECDF*n/(n+1) for the j-th column of x_train
        sorted_col = np.sort(x_train[:, j])
        ecdf = np.searchsorted(sorted_col, x_test[:, j], side="right") / (n_samples + 1)

        # Apply the transformation
        x_transf[:, j] = 1 / (1 - ecdf)

    return x_transf


def test_indep_radius_rest(
    X, y, ratio_ext, norm_func, random_state=np.random.randint(10**5)
):
    """Test the independence of the angular component and the\
    radius of extreme observations.

    This function performs a distance covariance test to assess the
    independence between the angular component and the radius of
    extreme observations. It uses a specified ratio to determine the
    threshold for extreme observations and performs the test for each
    ratio in `ratio_ext`.

    Parameters
    ----------
    X : array-like
        The input data matrix.
    y : array-like, optional
        The target vector. If None, only the angular component of `X` is
        considered.
    ratio_ext : float or list of float
        The ratio(s) used to determine the threshold for extreme observations.
    norm_func : callable
        A function that computes the norm of `X`.
    random_state : int, optional
        Seed for the random number generator. Default is a random integer.

    Returns
    -------
    tuple
        A tuple containing:
            - pvalues (ndarray): The p-values from the independence tests.
            - ratio_max (float): The maximum ratio for which the p-value is
                                 greater than 0.05.

    Notes
    -----
    - If `y` is None, only the angular component of `X` is considered.
    - For small sample sizes (< 100), a permutation-based test is used.
    - For larger sample sizes, an asymptotic test is used.
    - If no ratios satisfy the condition pvalues > 0.05, a warning is issued and ratio_max is set to 0.

    References
    ----------
    The method is inspired by the paper:
    
    Wan, P., & Davis, R. A. (2019). Threshold selection for multivariate heavy-tailed data. Extremes, 22(1), 131-166.
    """
    from dcor.independence import distance_covariance_test
    from dcor.independence import distance_correlation_t_test

    #
    if isinstance(ratio_ext, float):
        ratio_ext = [ratio_ext]
    norm_X = norm_func(X)
    Theta = X / norm_X.reshape(-1, 1)
    if y is None:
        Z = Theta
    else:
        Z = np.column_stack((Theta, y.reshape(-1, 1)))
    pvalues = []
    count = 0
    for ratio in ratio_ext:
        count += 173
        threshold = np.quantile(norm_X, 1 - ratio)
        id_extreme = norm_X >= threshold
        r_ext = norm_X[id_extreme]
        Z_ext = Z[id_extreme, :]
        if len(r_ext) < 100:
            # perform a distance covariance test with permutation-based
            # computation of the p-value
            test = distance_covariance_test(
                x=Z_ext,
                y=np.log(1 + r_ext).reshape(-1, 1),
                num_resamples=500,
                random_state=random_state + count,
            )
        else:
            # perform a distance covariance test with asymptotic p-value
            test = distance_correlation_t_test(
                x=Z_ext, y=np.log(1 + r_ext).reshape(-1, 1)
            )

        pvalues.append(test.pvalue)

    pvalues = np.array(pvalues)
    # Find indices where the condition is satisfied
    indices = np.where(pvalues > 0.05)[0]

    # Check if any indices satisfy the condition
    if len(indices) > 0:
        i_max = np.max(indices)
    else:
        warnings.warn("No indices satisfy the condition pvalues > 0.05.", UserWarning)

        i_max = 0

    ratio_max = ratio_ext[i_max]
    return pvalues, ratio_max


def plot_indep_radius_rest(pvalues, ratio_ext, ratio_max, n):
    """Plot the p-values from independence tests performed with \
    :func:`test_indep_radius_rest` against the number of extreme observations.

    This function creates a scatter plot of p-values from distance
    correlation tests against the number of extreme observations
    (k). It also highlights the selected k value based on a
    rule-of-thumb.

    Parameters
    ----------
    pvalues : array-like
        The p-values from the independence tests.
    ratio_ext : array-like
        The ratios used to determine the threshold for extreme observations.
    ratio_max : float
        The maximum ratio for which the p-value is greater than 0.05.
    n : int
        The total number of observations.

    Notes
    -----
    - The plot includes a secondary x-axis showing the ratio (k / n).
    - The selected k value is highlighted with a red line.
    - The plot title, labels, and legend are added for clarity.

    """
    kk = (ratio_ext * n).astype(int)
    k_max = int(ratio_max * n)
    fig, ax = plt.subplots()
    #    colors = ['green' if p > 0.05 else 'red' for p in pvalues]
    ax.scatter(kk, pvalues, c="black", label="distance correlation pvalues")
    ax.plot(
        k_max * np.ones(2),
        np.linspace(0, np.max(pvalues), num=2),
        c="red",
        label="selected k with distance covariance rule-of-thumb",
    )

    # Add a secondary x-axis with a different scale
    def ratio_transform(x):
        return x / n

    def inverse_ratio_transform(x):
        return x * n

    secax = ax.secondary_xaxis(
        "top", functions=(ratio_transform, inverse_ratio_transform)
    )
    secax.set_xlabel("Ratio (k / n)")
    ax.set_title("distance correlation test for extreme norm(X) Values vs Rest")
    ax.set_xlabel("k")
    ax.set_ylabel("p-value")
    ax.legend()
    plt.grid()
    plt.show()
